(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit CPoint;

interface

uses SysUtils, Utilities;

const
  // Die LAB Differenz ist der Menschlichen Farbempfindung sehr nahe
  LAB_DIFFERENCE = 1;

  // Schnell berrechneter RGB Farbunterschied nicht wirklich gut, aber da der
  // Farbunterschied unter umstnden so berechnet wurde ntzlich
  RGB_DIFFERENCE1 = 2;

  // Schnell berrechneter RGB Farbunterschied
  // der hchste unterschied in einem Kanal wird gewertet
  RGB_DIFFERENCE3 = 11;

  // Farbunterschied im Raum
  RGB_DIFFERENCE2 = 3;

  // unterschied des Farbtons
  HUE_DIFFERENCE = 4;

  // Unterschied der Farbsttigung
  SATURATION_DIFFERENCE = 5;

  // Unterschied der absoluten Helligkeit
  BRIGHTNESS_DIFFERENCE = 6;

  // Rotunterschied
  RED_DIFFERENCE = 7;

  // Grnunterschied
  GREEN_DIFFERENCE = 8;

  // Blauunterschied
  BLUE_DIFFERENCE = 9;

  // CMYK Color difference
  CMYK_DIFFERENCE = 10;

type
  TCPoint = record
    X, Y, Distance, Color: Integer;
    Foreground: Boolean;
    ColorDifferenceMode: Byte;

    function GetColorDifference(const AColor: Integer): Double;
  end;

implementation

uses untLogger;

{ TCPoint }

function TCPoint.GetColorDifference(const AColor: Integer): Double;
begin
  Result := 0;

  if AColor = Color then
    Exit;

  case ColorDifferenceMode of
    LAB_DIFFERENCE: Result := Utilities.GetColorDifference(AColor, Color);
    // xxx implement others
    RGB_DIFFERENCE2: Result := GetRGBColorDifference2(AColor, Color);
    // xxx
    BRIGHTNESS_DIFFERENCE: Result := GetBrightnessColorDifference(AColor, Color);

    else Logger.Warning('ColorDifferenceMode unimplemented (' + IntToStr(ColorDifferenceMode) + ')');
  end;
end;

end.
