unit GUI;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, untLogger, AntiCaptcha, PNGImage, Captcha,
  GifImg, ImgList;

type
  TfrmGUI = class(TForm)
    LogMemo: TMemo;
    cbMethods: TComboBox;
    lblMethods: TLabel;
    edtCaptcha: TButtonedEdit;
    Images: TImageList;
    OpenCaptcha: TOpenDialog;
    btnCheck: TButton;
    procedure btnCheckClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure edtCaptchaRightButtonClick(Sender: TObject);
  private
    FImages: array of TImage;
    FMethodDir: string;

    function CreateImage(X, Y, AWidth, AHeight: Integer): TImage;
  public
    procedure ShowImage(Img: TBitmap; Text: string = '');

    property MethodDir: string read FMethodDir write FMethodDir;
  end;

var
  frmGUI: TfrmGUI;

implementation

{$R *.dfm}

procedure TfrmGUI.FormCreate(Sender: TObject);
begin
  Logger.Control := LogMemo;
  FImages := nil;
end;

procedure TfrmGUI.FormDestroy(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to High(FImages) do
    FImages[i].Free;
  FImages := nil;
end;

procedure TfrmGUI.FormShow(Sender: TObject);
var
  Methods: TStringList;
begin
  Methods := TAntiCaptcha.GetMethods(FMethodDir);
  cbMethods.Items := Methods;
  if cbMethods.Items.Count > 0 then
    cbMethods.ItemIndex := 0
  else
    btnCheck.Enabled := False;
end;

function TfrmGUI.CreateImage(X, Y, AWidth, AHeight: Integer): TImage;
begin
  Result := TImage.Create(Self);
  Result.Parent := Self;
  Result.SetBounds(X, Y, AWidth, AHeight);
  Result.Center := True;
end;

procedure TfrmGUI.edtCaptchaRightButtonClick(Sender: TObject);
begin
  if OpenCaptcha.Execute then
    edtCaptcha.Text := OpenCaptcha.FileName;
end;

procedure TfrmGUI.ShowImage(Img: TBitmap; Text: string = '');
var
  X, Y: Integer;
begin
  if FImages = nil then
  begin
    SetLength(FImages, 1);
    FImages[0] := CreateImage(0, 0, Img.Width + 20, Img.Height + 20);
  end
  else
  begin
    X := FImages[High(FImages)].Left + FImages[High(FImages)].Width;
    Y := FImages[High(FImages)].Top;

    // Nchste Reihe
    if X + Img.Width + 20 > ClientWidth then
    begin
      X := 0;
      Inc(Y, FImages[High(FImages)].Height);
    end;

    SetLength(FImages, Length(FImages) + 1);
    FImages[High(FImages)] := CreateImage(X, Y, Img.Width + 20, Img.Height + 20);
  end;

  FImages[High(FImages)].Picture.Bitmap := Img;
  FImages[High(FImages)].Hint := Text;
  FImages[High(FImages)].ShowHint := True;

  Application.ProcessMessages;
end;

procedure TfrmGUI.btnCheckClick(Sender: TObject);
var
  AC: TAntiCaptcha;
begin
  if edtCaptcha.Text = '' then
    raise Exception.Create('Bitte ein Captcha auswhlen!');

  AC := TAntiCaptcha.Create(FMethodDir, cbMethods.Items[cbMethods.ItemIndex]);
  try
    AC.CheckCaptcha(edtCaptcha.Text);
  finally
    AC.Free;
  end;
end;

end.
