(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit LetterComperator;

interface

uses
  Windows, SysUtils, Letter, UTILITIES, Generics.Collections, Variants, Math;

const
  // IDs die Auskunft ber die Art der Erkennung geben
  PERFECT_MATCH = 1;            // Erkennung durch einen Perfectmatch Schwellwert
  ERROR = -1;                   // Keine Erkennung. Fehler!
  QUICKSCAN_MATCH = 2;          // Erkennung durch den quickscan
  QUICKSCAN_PERFECT_MATCH = 3;  // Perfect Match durch den Quickscan

  // Farbkonstanten fr die berlagerungsbilder
  BOTHCOLOR = $660099;
  BNACOLOR = $0000ff;
  ANBCOLOR = $ff0000;
  BNAFILTEREDCOLOR = $ccccff;
  ANBFILTEREDCOLOR = $ffcccc;

type
  TLetterComperator = class
  private
    FA, FB: TLetter;
    FBc: TArrayofInteger;
    FBothElements: TList<TList<Integer>>;
    FBothElementsNum: Integer;
    FCleftFaktor, FCoverageFaktorA, FCoverageFaktorAWeight: Double;
    FCoverageFaktorB, FCoverageFaktorBWeight: Double;
    FDetectionType: Integer;
    FDivider: Double;
    FElement: TList<Integer>;
    FElementGrid: T2DArrayofInteger;
    FErrorAWeight, FErrorBWeight: Double;
    FExtensionError, FHeightFaktor: Double;
    FImgOffset: TArrayofInteger;
    FIntersectionAHeightFaktor, FIntersectionAHeightWeight,
    FIntersectionAWidthFaktor, FIntersectionAWidthWeight,
    FIntersectionDimensionWeight: Double;
    FIntersectionGrid: T2DArrayofInteger;
    FIntersectionHeight: Integer;
    FIntersectionLetter: TLetter;
    FIntersectionStartX, FIntersectionStartY: Integer;
    FIntersectionWidth: Integer;
    FJas: TObject;   // TJACScript
    FLocalHeightPercent, FLocalWidthPercent: Double;
    FMinCleftSize: Integer;
    FOffset: TArrayofInteger;
    FOverlayNoiseSize: Integer;
    FOwner: TObject;          // TAntiCaptcha
    FPixelAll, FPixelANotB, FPixelBNotA, FPixelBoth: Integer;
    FPixelErrorA, FPixelErrorB, FPreScanDivider: Double;
    FPreScanFaktor, FPreScanFilter: Integer;
    FPreValityPercent: Double;

    FReliability: Double;
    FScanStepX, FScanStepY, FScanTime, FScanVarianceX, FScanVarianceY: Integer;
    FTmpCoverageFaktorA, FTmpCoverageFaktorB, FTmpErrorA, FTmpErrorB, FTmpErrorTotal: Double;
    FTmpHeightAFaktor, FTmpHeightFaktor: Double;
    FTmpPixelAButNotB, FTmpPixelBButNotA, FTmpPixelBoth: Integer;
    FTmpPreScanValue: Double;

    FTmpWidthAFaktor, FTmpWidthFaktor, FTotalPixelError, FValityPercent, FWidthFaktor: Double;

    function CoordinatesFromAToB(X, Y, XX, YY: Integer; Con: TArrayofInteger): TArrayofInteger;
    function GetElement(X, Y, XX, YY, Left, Top, PixelType: Integer; ElementGrid: T2DArrayofInteger; Counter: TList<Integer>): Integer;
    function GetPixelType(X, Y, XX, YY, Left, Top: Integer): Integer;
    function HasNeighbour(X, Y, XX, YY, Left, Top, PixelType: Integer): Integer;
    procedure Scan;
    function ScanIntersection(XX, YY, Left, Top, TmpIntersectionWidth, TmpIntersectionHeight: Integer): Double;
    function ScanPreIntersection(XX, YY, Left, Top, TmpIntersectionWidth, TmpIntersectionHeight: Integer): Double;

    function GetScanVarianceX: Integer;
    function GetScanVarianceY: Integer;
    function GetValityPercent: Double;
    procedure SetOwner(AOwner: TObject);   // TAntiCaptcha
  public
    // Gibt an ob ein Intersectionletter (schnittbilder) erstellt werden soll. Achtung langsam!
    CREATEINTERSECTIONLETTER: Boolean;

    IntersectionDimension: TArrayofInteger;
    TmpExtensionError: Double;

    constructor Create(AA, AB: TLetter);

    function GetDecodedValue: string;
    function GetDifference: TLetter;
    function GetIdentificationReliability: Double;
    function GetIntersection: TLetter;

    property A: TLetter read FA write FA;
    property B: TLetter read FB write FB;
    property BothElementsNum: Integer read FBothElementsNum write FBothElementsNum;
    property CoverageFaktorA: Double read FCoverageFaktorA write FCoverageFaktorA;
    property CoverageFaktorB: Double read FCoverageFaktorB write FCoverageFaktorB;
    property DetectionType: Integer read FDetectionType write FDetectionType;
    property ExtensionError: Double read FExtensionError write FExtensionError;
    property HeightFaktor: Double read FHeightFaktor write FHeightFaktor;
    property ImgOffset: TArrayofInteger read FImgOffset;
    property IntersectionAHeightFaktor: Double read FIntersectionAHeightFaktor write FIntersectionAHeightFaktor;
    property IntersectionAWidthFaktor: Double read FIntersectionAHeightFaktor write FIntersectionAHeightFaktor;
    property IntersectionHeight: Integer read FIntersectionHeight write FIntersectionHeight;
    property IntersectionLetter: TLetter read FIntersectionLetter write FIntersectionLetter;
    property IntersectionStartX: Integer read FIntersectionStartX write FIntersectionStartX;
    property IntersectionStartY: Integer read FIntersectionStartY write FIntersectionStartY;
    property IntersectionWidth: Integer read FIntersectionWidth write FIntersectionWidth;
    property LocalHeightPercent: Double read FLocalHeightPercent write FLocalHeightPercent;
    property Offset: TArrayofInteger read FOffset;
    property Owner: TObject read FOwner write SetOwner;
    property PixelANotB: Integer read FPixelANotB write FPixelANotB;
    property PixelBNotA: Integer read FPixelBNotA write FPixelBNotA;
    property PixelBoth: Integer read FPixelBoth write FPixelBoth;
    property PixelErrorA: Double read FPixelErrorA write FPixelErrorA;
    property PixelErrorB: Double read FPixelErrorB write FPixelErrorB;
    property PreValityPercent: Double read FPreValityPercent write FPreValityPercent;
    property RealValityValue: Double read FValityPercent write FValityPercent;
    property Reliability: Double read FReliability write FReliability;
    property ScanTime: Integer read FScanTime write FScanTime;
    property ScanVarianceX: Integer read GetScanVarianceX;
    property ScanVarianceY: Integer read GetScanVarianceY;
    property TotalPixelError: Double read FTotalPixelError write FTotalPixelError;
    property ValityPercent: Double read GetValityPercent write FValityPercent;
    property WidthFaktor: Double read FWidthFaktor;

    function IsCreateIntersectionLetter: Boolean;

    procedure Run;

    procedure SetScanVariance(X, Y: Integer);

    function ToString: string; override;
  end;

implementation

uses AntiCaptcha, JACScript, PixelGrid, untLogger;

constructor TLetterComperator.Create(AA, AB: TLetter);
begin
  FA := AA;
  FB := AB;

  CREATEINTERSECTIONLETTER := False;

  SetLength(FBc, 2);
  FBothElements := TList<TList<Integer>>.Create;
  FCoverageFaktorA := 0;
  FCoverageFaktorB := 0;

  FExtensionError := 0.0;
  FHeightFaktor := 0;
  FIntersectionHeight := 0;
  FIntersectionLetter := TLetter.Create;
  FIntersectionStartX := 0;
  FIntersectionStartY := 0;
  FIntersectionWidth := 0;
  FPixelANotB := 0;
  FPixelBNotA := 0;
  FPixelBoth := 0;
  FPixelErrorA := 0;
  FPixelErrorB := 0;

  FScanTime := -1;
  FScanVarianceX := -1;
  FScanVarianceY := -1;
  TmpExtensionError := 0.0;
  FTmpPixelBoth := 0;

  FTotalPixelError := 0;
  FValityPercent := 10000.0;
  FWidthFaktor := 0;
end;

function TLetterComperator.CoordinatesFromAToB(X, Y, XX, YY: Integer; Con: TArrayofInteger): TArrayofInteger;
begin
  Con[0] := X - XX;
  Con[1] := Y - YY;
  Result := Con;
end;

function TLetterComperator.GetDecodedValue: string;
{ Gibt den Decoded Value von b zurck }
begin
  if (FB = nil) or (Length(FB.DecodedValue) <> 1) then
    Exit('-');

  Result := FB.DecodedValue;
end;

function TLetterComperator.GetDifference: TLetter;
var
  XX, YY, Left, Top, TmpIntersectionWidth, TmpIntersectionHeight: Integer;
  g: T2DArrayofInteger;
  x, y, PixelType: Integer;
  l: TStatic01ArrayofInt;
begin
  try
    XX := FImgOffset[0];
    YY := FImgOffset[1];
    Left := FOffset[0];
    top := FOffset[1];
    TmpIntersectionWidth := IntersectionDimension[0];
    TmpIntersectionHeight := IntersectionDimension[1];

    SetLength(g, TmpIntersectionWidth, TmpIntersectionHeight);

    for x := 0 to TmpIntersectionWidth - 1 do
      for y := 0 to TmpIntersectionHeight - 1 do
      begin
        g[x][y] := FA.GetMaxPixelValue;
        PixelType := GetPixelType(X, Y, XX, YY, Left, top);

        case PixelType of
          1:
          begin
            if HasNeighbour(X, Y, XX, YY, Left, top, PixelType) > FOverlayNoiseSize then
              // g[x][y] = 0xff0000;
            else
            begin
              // g[x][y] = 0xff0000;
            end;

            g[x][y] := 0;
          end;

          2:
          begin
            if HasNeighbour(X, Y, XX, YY, Left, top, PixelType) > FOverlayNoiseSize then
              g[x][y] := 0;
              // g[x][y] = 0x00ff00;
          end;
        end;
      end;

    Result := TLetter(FA.CreateLetter);
    l := FA.Location;
    l[0] := l[0] + Left;
    l[1] := l[1] + top;
    Result.Location := l;
    Result.Grid := g;
    Result.Clean;
  except
    Result := FA;
  end;
end;

function TLetterComperator.GetElement(x: Integer; y: Integer; XX: Integer;
  YY: Integer; Left: Integer; top: Integer; PixelType: Integer;
  ElementGrid: T2DArrayofInteger; Counter: TList<Integer>): Integer;
var
  pt: Integer;
begin
  if (x < 0) or (y < 0) or (x >= Length(ElementGrid)) or (Length(ElementGrid) = 0) or (y >= Length(elementGrid[0])) then
    Exit(0);

  if (ElementGrid[x][y] <> 0) then
    Exit(0);

  pt := GetPixelType(x, y, XX, YY, Left, top);

  if pt = PixelType then
  begin
    Counter.Add(PixelType);
    ElementGrid[x][y] := PixelType + 100;
    GetElement(x - 1, y, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x - 1, y - 1, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x, y - 1, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x + 1, y - 1, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x + 1, y, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x + 1, y + 1, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x, y + 1, XX, YY, Left, top, pt, ElementGrid, Counter);
    GetElement(x - 1, y + 1, XX, YY, Left, top, pt, ElementGrid, Counter);
  end;

  Result := 0;
end;

function TLetterComperator.GetIdentificationReliability: Double;
begin
  Result := ValityPercent - FReliability;
end;

function TLetterComperator.GetIntersection: TLetter;
var
  Left, top, TmpIntersectionWidth, TmpIntersectionHeight: Integer;
  x, y: Integer;
  g: T2DArrayofInteger;
  l: TStatic01ArrayofInt;
begin
  try
    // XX := Self.imgOffset[0];
    // YY := Self.imgOffset[1];
    Left := FOffset[0];
    top := FOffset[1];
    TmpIntersectionWidth := intersectionDimension[0];
    TmpIntersectionHeight := intersectionDimension[1];

    SetLength(g, TmpIntersectionWidth, TmpIntersectionHeight);

    for x := 0 to TmpIntersectionWidth - 1 do
      for y := 0 to TmpIntersectionHeight - 1 do
        g[x][y] := FA.GetPixelValue(x + Left, y + top);

    Result := TLetter(FA.createLetter);
    l := FA.Location;
    l[0] := l[0] + Left;
    l[1] := l[1] + top;
    Result.Location := l;
    Result.Grid := g;
    Result.clean;
  except
    logger.warning('Exception caught @ TLetterComperator.getIntersection: ' + Exception(ExceptObject).Message);
    Result := FA;
  end;
end;

function TLetterComperator.GetPixelType(x, y, XX, YY, Left, top: Integer): Integer;
var
  va, vb: Integer;
begin
  va := FA.GetPixelValue(x + Left, y + top);
  FBc := CoordinatesFromAToB(x + Left, y + top, XX, YY, FBc);
  vb := FB.GetPixelValue(FBc[0], FBc[1]);
  if (va < 0) or (vb < 0) then
    Exit(-2);
  if (va = 0) and (vb = 0) then
    Exit(0)
  else
    if vb = 0 then
      Exit(1)
    else
      if va = 0 then
        Exit(2);

  Result := -1;
end;

function TLetterComperator.GetScanVarianceX: Integer;
begin
  if (FScanVarianceX >= 0) then
    Exit(FScanVarianceX);

  Result := Integer(TJACScript(FJas).GetParam('scanVarianceX'));
end;

function TLetterComperator.GetScanVarianceY: Integer;
begin
  if (FScanVarianceX >= 0) then
    Exit(FScanVarianceY);

  Result := Integer(TJACScript(FJas).GetParam('scanVarianceY'));
end;

function TLetterComperator.GetValityPercent: Double;
begin
  Result := Min(100.0, FValityPercent);
end;

function TLetterComperator.HasNeighbour(X, Y, XX, YY, Left, Top, PixelType: Integer): Integer;
var
  Faktor, xt, yt: Integer;
begin
  Result := 0;
  Faktor := 1;
  for xt := -Faktor to Faktor do
    for yt := -Faktor to Faktor do
    begin
      if (xt = 0) and (yt = 0) then
        Continue;

      if (GetPixelType(x + xt, y + yt, XX, YY, Left, top) = PixelType) then
        Inc(Result);
    end;
end;

function TLetterComperator.IsCreateIntersectionLetter: Boolean;
begin
  Result := CREATEINTERSECTIONLETTER;
end;

procedure TLetterComperator.Run;
{ Fhrt den Vergleichsvorgang aus }
begin
  Scan;
end;

procedure TLetterComperator.Scan;
{ Scan ist die eigentliche Vergleichsfunktion. a und b werden dabei
  gegeneinander verschoben. }
var
  StartTime: Cardinal;
  BestValue, Value: Double;
  VX, VY: Integer;
  ScanXFrom, ScanXTo, ScanYFrom, ScanYTo, Tmp: Integer;
  TmpIntersection: TLetter;
  Left, Right, Top, Bottom, TmpIntersectionWidth, TmpIntersectionHeight: Integer;
  XX, YY: Integer;
begin
  StartTime := GetTickCount;
  BestValue := 20000.0;
  PreValityPercent := 20000.0;
  FTmpPreScanValue := 20000.0;
  // Logger.Info(b.DecodedValue + '----');

  // Scanvarianzen geben an, wieviel beim verschieben ber die Grenzen
  // geschoben wird. Groe Werte brauchen CPU!
  VX := ScanVarianceX;
  VY := ScanVarianceY;
  VX := Min(VX, FB.Width);
  VY := Min(VY, FB.Height);
  ScanXFrom := -VX;
  ScanXTo := FA.Width - FB.Width + VX;
  ScanYFrom := -VY;
  ScanYTo := FA.Height - FB.Height + VY;

  if ScanXTo < ScanXFrom then
  begin
    Tmp := ScanXTo;
    ScanXTo := ScanXFrom;
    ScanXFrom := Tmp;
  end;

  if ScanYTo < ScanYFrom then
  begin
    Tmp := ScanYTo;
    ScanYTo := ScanYFrom;
    ScanYFrom := Tmp;
  end;

  if IsCreateIntersectionLetter then
    TmpIntersection := TLetter.Create;

  XX := GetJumperStart(ScanXFrom, ScanXTo);
  while CheckJumper(XX, ScanXFrom, ScanXTo) do
  begin
    YY := GetJumperStart(ScanYFrom, ScanYTo);
    while CheckJumper(YY, ScanYFrom, ScanYTo) do
    begin
      // Offsets
      Left := Max(0, XX);
      Right := Min(XX + FB.Width, FA.Width);
      Top := Max(0, YY);
      Bottom := Min(YY + FB.Height, FA.Height);
      // intersection ^= ausschnitt
      TmpIntersectionWidth := Right - Left;
      TmpIntersectionHeight := Bottom - Top;
      if (TmpIntersectionWidth <= 0) or (TmpIntersectionHeight <= 0) then
      begin
        // Logger.Warning('Scannvarianzen zu gro: ' +
        // Format('%d/%d', [TmpIntersectionWidth, TmpIntersectionHeight]));
        YY := NextJump(YY, ScanYFrom, ScanYTo, 1);

        Continue;
      end;

      if IsCreateIntersectionLetter then
      begin
        TmpIntersection := TLetter.Create;
        TmpIntersection.Owner := Owner;
        SetLength(FIntersectionGrid, TmpIntersectionWidth, TmpIntersectionHeight);
        TmpIntersection.Grid := FIntersectionGrid;
      end
      else
        TmpIntersection := nil;
      //
      if FPreScanFilter > 0 then
      begin
        FTmpPreScanValue := ScanPreIntersection(XX, YY, Left, Top, TmpIntersectionWidth, TmpIntersectionHeight);

        if FTmpPreScanValue > FPreScanFilter then
        begin
          if FPreValityPercent > FTmpPreScanValue then
            FPreValityPercent := FTmpPreScanValue;

          YY := NextJump(YY, ScanYFrom, ScanYTo, 1);
          Continue;
        end;
      end;

      Value := ScanIntersection(XX, YY, Left, Top, TmpIntersectionWidth, TmpIntersectionHeight);
      if Value < BestValue then
      begin
        BestValue := Value;
        FExtensionError := TmpExtensionError;
        FValityPercent := Value;
        FHeightFaktor := FTmpHeightFaktor;
        FWidthFaktor := FTmpWidthFaktor;
        FIntersectionHeight := TmpIntersectionHeight;
        FIntersectionWidth := TmpIntersectionWidth;
        FIntersectionStartX := XX;
        FIntersectionStartY := YY;
        FBothElementsNum := FBothElements.Count;
        FCoverageFaktorA := FTmpCoverageFaktorA;
        FCoverageFaktorB := FTmpCoverageFaktorB;
        FPixelErrorA := FTmpErrorA;
        FPreValityPercent := FTmpPreScanValue;
        FPixelErrorB := FTmpErrorB;
        FPixelANotB := FTmpPixelAButNotB;
        FPixelBNotA := FTmpPixelBButNotA;
        FPixelBoth := FTmpPixelBoth;
        FIntersectionAHeightFaktor := 0.0;
        FIntersectionAWidthFaktor := 0.0;
        FIntersectionAHeightFaktor := FTmpHeightAFaktor;
        FIntersectionAWidthFaktor := FTmpWidthAFaktor;

        FTotalPixelError := FTmpErrorTotal;

        if IsCreateIntersectionLetter then
          FIntersectionLetter := TmpIntersection;
      end;

      YY := NextJump(YY, ScanYFrom, ScanYTo, 1);
    end;
    XX := NextJump(XX, ScanXFrom, ScanXTo, 1);
  end;

  ScanTime := GetTickCount - StartTime;
end;

function TLetterComperator.ScanIntersection(XX, YY, Left, top,
  TmpIntersectionWidth, TmpIntersectionHeight: Integer): Double;
var
  TmpError: Double;
  x, y, PixelType: Integer;
begin
  SetLength(FOffset, 2);
  FOffset[0] := Left;
  FOffset[1] := Top;
  SetLength(FImgOffset, 2);
  FImgOffset[0] := XX;
  FImgOffset[1] := YY;
  SetLength(IntersectionDimension, 2);
  IntersectionDimension[0] := TmpIntersectionWidth;
  IntersectionDimension[1] := TmpIntersectionHeight;

  FPixelAll := 0;
  FTmpPixelBButNotA := 0;
  FTmpPixelAButNotB := 0;
  FTmpPixelBoth := 0;
  FTmpCoverageFaktorA := 0;
  FTmpCoverageFaktorB := 0;
  //starter := GetTickCount;
  FBothElements.Clear;
  SetLength(FElementGrid, TmpIntersectionWidth, TmpIntersectionHeight);

  x := 0;
  while x < TmpIntersectionWidth do
  begin
    y := 0;
    while y < TmpIntersectionHeight do
    begin
      PixelType := GetPixelType(x, y, XX, YY, Left, top);
      Inc(FPixelAll);

      case PixelType of
        0:
        begin
          if isCreateIntersectionLetter then
            FIntersectionGrid[x][y] := BOTHCOLOR;

          if FCleftFaktor > 0 then
          begin
            FElement := TList<Integer>.Create;
            getElement(x, y, XX, YY, Left, top, PixelType, FElementGrid, FElement);
            if FElement.Count > FMinCleftSize then
              FBothElements.Add(FElement);

            Inc(FTmpPixelBoth);
          end;
        end;

        1:
        begin
          if (FOverlayNoiseSize <= 0) or (hasNeighbour(x, y, XX, YY, Left, top, PixelType) > FOverlayNoiseSize) then
          begin
            Inc(FTmpPixelBButNotA);
            if IsCreateIntersectionLetter then
              FIntersectionGrid[x][y] := BNACOLOR;
          end
          else
            if (isCreateIntersectionLetter()) then
              FIntersectionGrid[x][y] := BNAFILTEREDCOLOR;
        end;

        2:
        begin
          if (FOverlayNoiseSize <= 0) or (hasNeighbour(x, y, XX, YY, Left, top, PixelType) > FOverlayNoiseSize) then
          begin
            Inc(FTmpPixelBButNotA);
            if isCreateIntersectionLetter then
              FIntersectionGrid[x][y] := ANBCOLOR;
          end
          else
            if (isCreateIntersectionLetter()) then
              FIntersectionGrid[x][y] := ANBFILTEREDCOLOR;
        end;

        else
          if (isCreateIntersectionLetter) then
            FIntersectionGrid[x][y] := $ffffff;

      end;

      Inc(y, FScanStepY);
    end;

    Inc(x, FScanStepX);
  end;

  // logger.info('Scanner: ' + IntToStr(GetTickCount - starter));
  // if (LowerCase(getDecodedValue) = 'v') and (getBothElementsNum = 3) then
  //   logger.info('JJJ');

  if (FPixelAll > 0) and ((FBothElements.Count > 0) or (FCleftFaktor = 0)) then
  begin
    if (FTmpPixelAButNotB <> 0) and (FTmpPixelBoth + FTmpPixelAButNotB <> 0) then
      FTmpErrorA := FTmpPixelAButNotB / (FTmpPixelBoth + FTmpPixelAButNotB)
    else
      FTmpErrorA := 0;

    if (FTmpPixelBButNotA <> 0) and (FTmpPixelBoth + FTmpPixelBButNotA <> 0) then
      FTmpErrorB := FTmpPixelBButNotA / (FTmpPixelBButNotA + FTmpPixelBoth)
    else
      FTmpErrorB := 0;

    FTmpErrorTotal := FTmpErrorA * FErrorAWeight + FTmpErrorB * FErrorbWeight;

    if (FScanStepX * FScanStepY) <> 0 then
    begin
      FTmpCoverageFaktorA := 1.0 - FTmpPixelBoth / (FA.GetElementPixel / (FScanStepX * FScanStepY));
      FTmpCoverageFaktorB := 1.0 - FTmpPixelBoth / (FB.GetElementPixel / (FScanStepX * FScanStepY));
    end
    else
    begin
      FTmpCoverageFaktorA := 0;
      FTmpCoverageFaktorB := 0;
    end;

    FLocalHeightPercent := TmpIntersectionHeight / FB.Height;
    FLocalWidthPercent := TmpIntersectionWidth / FB.Width;

    FTmpHeightFaktor := Power(1.0 - FLocalHeightPercent, 2);
    FTmpWidthFaktor := Power(1.0 - FlocalWidthPercent, 2);
    FTmpHeightAFaktor := 1.0 - TmpIntersectionHeight / FA.Height;
    FTmpWidthAFaktor := 1.0 - TmpIntersectionWidth / FA.Width;
    // logger.info(TmpIntersectionWidth+ '/'+a.getWidth+' =
    // '+localWidthPercent+' --> '+TmpWidthFaktor);

    TmpError := FTmpErrorTotal;
    TmpError := TmpError + Min(1.0, FTmpCoverageFaktorA) * FCoverageFaktorAWeight;
    TmpError := TmpError + Min(1.0, FTmpCoverageFaktorB) * FCoverageFaktorBWeight;
    TmpError := TmpError + Min(1.0, FTmpHeightFaktor) * FIntersectionDimensionWeight;
    TmpError := TmpError + Min(1.0, FTmpWidthFaktor) * FIntersectionDimensionWeight;
    TmpError := TmpError + Min(1.0, FTmpHeightAFaktor) * FIntersectionAHeightWeight;
    TmpError := TmpError + Min(1.0, FTmpWidthAFaktor) * FIntersectionAWidthWeight;

    if FBothElements.Count > 0 then
      TmpError := TmpError + (FBothElements.Count - 1) * FCleftFaktor;

    TmpExtensionError := 0.0;

    TmpError := TmpError + TmpExtensionError;
    if FDivider = 0 then
      TmpError := 10000
    else
      TmpError := TmpError / FDivider;
    // TmpError := Min(1.0, TmpError);
    // logger.info(pixelBoth+'_'+(TmpIntersectionHeight *
    // TmpIntersectionWidth));
    if (FTmpPixelBoth * Double(TJACScript(FJas).GetParam('inverseFontWeight')) < TmpIntersectionHeight * TmpIntersectionWidth) then
    begin
      TmpError := 10000.0 / 100.0;
      FTmpErrorA := TmpError;
      FTmpErrorB := TmpError;
      FTmpErrorTotal := TmpError;
    end;

    Result := 100.0 * TmpError;
  end
  else
    Result := 10000;
end;

function TLetterComperator.ScanPreIntersection(XX, YY, Left, top,
  TmpIntersectionWidth, TmpIntersectionHeight: Integer): Double;
var
  TmpError: Double;
  yStep, x, y, PixelType: Integer;
begin
  yStep := Max(1, TmpIntersectionHeight div (FPreScanFaktor + 1));

  FTmpPixelBoth := 0;
  FTmpPixelBButNotA := 0;
  FTmpPixelAButNotB := 0;
  FpixelAll := 0;

  y := yStep;
  while y <= TmpIntersectionHeight - yStep do
  begin
    for x := 0 to TmpIntersectionWidth - 1 do
    begin
      Inc(FpixelAll);

      PixelType := getPixelType(x, y, XX, YY, Left, top);
      case PixelType of
        0: Inc(FTmpPixelBoth);
        1: Inc(FTmpPixelBButNotA);
        2: Inc(FTmpPixelAButNotB);
      end;
    end;

    y := y + yStep;
  end;

  // logger.info((GetTickCount - startTime)+ 'intersection
  // scantime: '+(GetTickCount - starter2));
  if FpixelAll > 0 then
  begin
    if FTmpPixelBoth + FTmpPixelAButNotB = 0 then
      FTmpErrorA := 10000
    else
      FTmpErrorA := FTmpPixelAButNotB / (FTmpPixelBoth + FTmpPixelAButNotB);

    if FTmpPixelBButNotA + FTmpPixelBoth = 0 then
      FTmpErrorB := 10000
    else
      FTmpErrorB := FTmpPixelBButNotA / (FTmpPixelBButNotA + FTmpPixelBoth);

    FTmpErrorTotal := FTmpErrorA * FErrorAWeight + FTmpErrorB * FErrorbWeight;
    FLocalHeightPercent := TmpIntersectionHeight / FB.Height;
    FLocalWidthPercent := TmpIntersectionWidth / FB.Width;
    FTmpHeightFaktor := Power(1.0 - FLocalHeightPercent, 2);
    FTmpWidthFaktor := Power(1.0 - FlocalWidthPercent, 2);

    // TmpHeightAFaktor := Power(1.0 - TmpIntersectionHeight / a.getHeight, 2);
    // TmpWidthAFaktor := Power(1.0 - TmpIntersectionWidth / a.getWidth, 2);
    TmpError := FTmpErrorTotal;
    TmpError := TmpError + Min(1.0, FTmpHeightFaktor) * FintersectionDimensionWeight;
    TmpError := TmpError + Min(1.0, FTmpWidthFaktor) * FintersectionDimensionWeight;
    TmpError := TmpError + Min(1.0, FTmpHeightAFaktor) * FintersectionAHeightWeight;
    TmpError := TmpError + Min(1.0, FTmpWidthAFaktor) * FintersectionAWidthWeight;

    if FprescanDivider <> 0 then
      TmpError := TmpError / FprescanDivider
    else
      TmpError := 100;

    TmpError := TmpError * 1.2;

    Result := (100 * TmpError);
  end
  else
    Result := 100;
end;

procedure TLetterComperator.SetOwner(AOwner: TObject);
begin
  FOwner := AOwner;
  FJas := TAntiCaptcha(FOwner).Script;
  FErrorAWeight := Double(TJACScript(FJas).GetParam('errorAWeight'));
  FErrorbWeight := Double(TJACScript(FJas).GetParam('errorBWeight'));
  FCoverageFaktorAWeight := Double(TJACScript(FJas).GetParam('coverageFaktorAWeight'));
  FCoverageFaktorBWeight := Double(TJACScript(FJas).GetParam('coverageFaktorBWeight'));
  FIntersectionDimensionWeight := Double(TJACScript(FJas).GetParam('intersectionDimensionWeight'));
  FIntersectionAHeightWeight := Double(TJACScript(FJas).GetParam('intersectionAHeightWeight'));
  FIntersectionAWidthWeight := Double(TJACScript(FJas).GetParam('intersectionAWidthWeight'));
  FCleftFaktor := Double(TJACScript(FJas).GetParam('CleftFaktor'));
  FMinCleftSize := Integer(TJACScript(FJas).GetParam('minCleftSize'));
  FPreScanFilter := Integer(TJACScript(FJas).GetParam('preScanFilter'));
  FPreScanFaktor := Integer(TJACScript(FJas).GetParam('preScanFaktor'));
  FOverlayNoiseSize := Integer(TJACScript(FJas).GetParam('overlayNoiseSize'));
  FScanStepX := Integer(TJACScript(FJas).GetParam('scanstepx'));
  FScanStepY := Integer(TJACScript(FJas).GetParam('scanstepy'));
  FPrescanDivider := Double(TJACScript(FJas).GetParam('prescandivider'));
  FDivider := Double(TJACScript(FJas).GetParam('divider'));
  // xxx
end;

procedure TLetterComperator.SetScanVariance(X, Y: Integer);
begin
  FScanVarianceX := X;
  FScanVarianceY := Y;
end;

function TLetterComperator.ToString: string;
begin
  // xxx
end;

end.
