(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit PixelGrid;

interface

uses Windows, SysUtils, Math, CSTools, UTILITIES, Graphics, Classes, jpeg,
     PixelObject, Generics.Collections, Generics.Defaults, untLogger;

type
  TStatic01ArrayofInt = array[0..1] of Integer;

  TPixelGrid = class
  private
    procedure GetObject(X, Y: Integer; TmpGrid: T2DArrayofInteger; PObject: TPixelObject);
  protected
    FTmpGrid: T2DArrayofInteger;
    FOwner: TObject;    // TAntiCaptcha
    FGrid: T2DArrayofInteger;
    FPixel: TArrayofInteger;
    FLocation: TStatic01ArrayofInt;

    function GetHeight: Integer;
    function GetWidth: Integer;

    function GetColorObjects(letterNum: Integer): TArrayofPixelObject;
  public
    class procedure FillLetter(Lt: TObject);
    class function GetDimension(AGrid: T2DArrayofInteger): TArrayofInteger;
    class function GetGridCopy(AGrid: T2DArrayofInteger): T2DArrayofInteger; overload;
    class function GetGridHeight(AGrid: T2DArrayofInteger): Integer;
    class function GetGridWidth(AGrid: T2DArrayofInteger): Integer;
    class function GetMaxPixelValue(AOwner: TObject): Integer; overload;
    class function GetPixelValue(X, Y: Integer; AGrid: T2DArrayofInteger): Integer; overload;
    class function GetValityPercent(Value: Integer; AOwner: TObject): Integer;
    class procedure RecFill(P: TPixelObject; Lt: TObject; X, Y: Integer; Tmp: T2DArrayofInteger; i: Integer);
    class procedure SetPixelValue(X, Y: Integer; LocalGrid: T2DArrayofInteger; Value: Integer); overload;

    constructor Create(AWidth, AHeight: Integer);

    procedure BlurIt(Faktor: Integer);
    function Clean: Boolean;
    procedure CleanBackgroundByColor(Avg: Integer);
    procedure CleanBackgroundBySample(PX, PY, AWidth, AHeight: Integer);
    procedure ColorObject(PObject: TPixelObject; Color: Integer);
    function CreateLetter: TObject;
    procedure Crop(LeftPadding, TopPadding, RightPadding, BottomPadding: Integer);
    procedure DesinX(Max, Omega, Phi: Double);
    procedure DesinY(Max, Omega, Phi: Double);
    function GetAverage: Integer; overload;
    function GetAverage(PX, PY, AWidth, AHeight: Integer): Integer; overload;
    function GetAverageWithoutPoint(PX, PY, AWidth, AHeight: Integer): Integer;
    function GetDim: string;
    function GetFullImage: TBitmap;
    function GetGridCopy: T2DArrayofInteger; overload;
    function GetImage: TBitmap; overload;
    function GetImage(Faktor: Integer): TBitmap; overload;
    function GetMaxPixelValue: Integer; overload;
    function GetMaxPixelValue(Faktor: Double): Integer; overload;
    function GetObjects(Contrast, ObjectContrast: Double): TList<TPixelObject>;
    function GetPixel: TArrayofInteger;
    function GetPixelValue(X, Y: Integer): Integer; overload;
    function GetString: string;
    procedure Invert;
    function IsElement(Value, Avg: Integer): Boolean;
    procedure Normalize; overload;
    procedure Normalize(Multi: Double); overload;
    procedure Normalize(Multi, CutMax, CutMin: Double); overload;
    procedure PrintGrid;
    procedure ReduceBlackNoise(Faktor: Integer); overload;
    procedure ReduceBlackNoise(Faktor: Integer; Contrast: Double); overload;
    procedure ReduceWhiteNoise(Faktor: Integer); overload;
    procedure ReduceWhiteNoise(Faktor: Integer; Contrast: Double); overload;
    procedure RemoveBridges(Pixels: Integer; Middle: Double);
    procedure RemoveObjectFromGrid(PObject: TPixelObject);
    procedure RemoveSmallObjects(Contrast, ObjectContrast: Double); overload;
    procedure RemoveSmallObjects(Contrast, ObjectContrast: Double; MaxSize: Integer); overload;
    procedure SampleDown(Faktor: Integer);
    procedure SaveImageAsJpeg(FileName: string);
    procedure SetOrgGrid(AGrid: T2DArrayofInteger);
    procedure SetPixel(APixel: TArrayofInteger);
    procedure SetPixelValue(X, Y, Value: Integer); overload;
    procedure TestColor;
    procedure ToBlackAndWhite; overload;
    procedure ToBlackAndWhite(Contrast: Double); overload;
    function ToHsbColorString: string;

    property Height: Integer read GetHeight;
    property Width: Integer read GetWidth;
    property Grid: T2DArrayofInteger read FGrid write FGrid;
    property Location: TStatic01ArrayofInt read FLocation write FLocation;
    property Owner: TObject read FOwner write FOwner;
  end;

implementation

uses AntiCaptcha, JACScript, Letter {$IFDEF DEBUG}, GUI{$ENDIF};

class procedure TPixelGrid.FillLetter(Lt: TObject);
const
  LIMIT = 200;
var
  L: TLetter;
  Tmp: T2DArrayofInteger;
  X, Y: Integer;
  PO: TPixelObject;
begin
  L := TLetter(Lt);

  SetLength(Tmp, L.Width, L.Height);

  for X := 0 to L.Width - 1 do
    for Y := 0 to L.Height - 1 do
      if (L.Grid[X][Y] > LIMIT) and (Tmp[X][Y] <> 1) then
      begin
        PO := TPixelObject.Create(L);
        RecFill(PO, L, X, Y, Tmp, 0);
        if (PO.Bordered) and (PO.Size < 60) then
          L.FillWithObject(PO, 0);
      end;
end;

class function TPixelGrid.GetDimension(AGrid: T2DArrayofInteger): TArrayofInteger;
var
  TopLines, BottomLines, LeftLines, RightLines: Integer;
  Width, Height, X, Y: Integer;
  Broken: Boolean;
begin
  TopLines := 0;
  BottomLines := 0;
  LeftLines := 0;
  RightLines := 0;

  Width := Length(AGrid);
  Height := Length(AGrid[0]);

  Broken := False;

  // row:
  for X := 0 to Width - 1 do
  begin
    for Y := 0 to Height - 1 do
      if AGrid[X][Y] = 0 then
      begin
        Broken := True;
        Break;
      end;

    if Broken then
      Break;

    Inc(LeftLines);
  end;

  Broken := False;    // zurcksetzen

  // row:
  for X := Width - 1 downto 0 do
  begin
    for Y := 0 to Height - 1 do
      if AGrid[X][Y] = 0 then
      begin
        Broken := True;
        Break;
      end;

    if Broken then
      Break;

    Inc(RightLines);
  end;

  if (LeftLines >= Width) or (Width - RightLines > Width) then
  begin
    SetLength(Result, 2);
    Result[0] := 0;
    Result[1] := 0;
    Exit;
  end;

  Broken := False;    // zurcksetzen

  // line:
  for Y := 0 to Height - 1 do
  begin
    for X := LeftLines to (Width - RightLines) - 1 do
      if AGrid[X][Y] = 0 then
      begin
        Broken := True;
        Break;
      end;

    if Broken then
      Break;

    Inc(TopLines);
  end;

  // line:
  for Y := Height - 1 downto 0 do
  begin
    for X := LeftLines to (Width - RightLines) - 1 do
      if AGrid[X][Y] = 0 then
      begin
        Broken := True;
        Break;
      end;

    if Broken then
      Break;

    Inc(BottomLines);
  end;

  if (Width - LeftLines - RightLines < 0) or (Height - TopLines - BottomLines < 0) then
  begin
    SetLength(Result, 2);
    Result[0] := 0;
    Result[1] := 0;
    Exit;
  end;

  SetLength(Result, 2);
  Result[0] := Width - LeftLines - RightLines;
  Result[1] := Height - TopLines - BottomLines;
end;

class function TPixelGrid.GetGridCopy(AGrid: T2DArrayofInteger): T2DArrayofInteger;
var
  X, Y: Integer;
begin
  if AGrid = nil then
    Exit(nil);

  SetLength(Result, Length(AGrid), Length(AGrid[0]));
  for X := 0 to High(AGrid) do
    for Y := 0 to High(AGrid[0]) do
      Result[X][Y] := AGrid[X][Y];
end;

class function TPixelGrid.GetGridHeight(AGrid: T2DArrayofInteger): Integer;
begin
  if AGrid = nil then
    Exit(0);

  Result := Length(AGrid[0]);
end;

class function TPixelGrid.GetGridWidth(AGrid: T2DArrayofInteger): Integer;
begin
  Result := Length(AGrid);
end;

class function TPixelGrid.GetMaxPixelValue(AOwner: TObject): Integer;
begin
  Result := (AOwner as TAntiCaptcha).Script.ColorFaktor - 1;
end;

class function TPixelGrid.GetPixelValue(X, Y: Integer; AGrid: T2DArrayofInteger): Integer;
begin
  if (X < 0) or (X >= Length(AGrid)) then
    Exit(-1);

  if (Y < 0) or (Length(AGrid) < 1) or (Y >= Length(AGrid[0])) then
    Exit(-1);

  Result := AGrid[X][Y];
end;

class function TPixelGrid.GetValityPercent(Value: Integer; AOwner: TObject): Integer;
begin
  if Value < 0 then
    Exit(100);

  Result := Trunc((100.0 * Value) / GetMaxPixelValue(AOwner));
end;

class procedure TPixelGrid.RecFill(P: TPixelObject; Lt: TObject; X, Y: Integer; Tmp: T2DArrayofInteger; i: Integer);
var
  L: TLetter;
begin
  Inc(i);

  L := TLetter(Lt);

  if (X >= 0) and (Y >= 0) and (X < L.Width) and (Y < L.Height) and (L.Grid[X][Y] > 200) and (Tmp[X][Y] <> 1) then
  begin
    if (X = 0) or (Y = 0) or (X = L.Width - 1) or (Y = L.Height - 1) then
      P.Bordered := False;

    P.Add(X, Y, $FF0000);
    Tmp[X][Y] := 1;

    RecFill(P, L, X - 1, Y, Tmp, i);
    RecFill(P, L, X, Y - 1, Tmp, i);
    RecFill(P, L, X + 1, Y, Tmp, i);
    RecFill(P, L, X, Y + 1, Tmp, i);
  end;
end;

class procedure TPixelGrid.SetPixelValue(X, Y: Integer; LocalGrid: T2DArrayofInteger; Value: Integer);
begin
  try
    LocalGrid[X][Y] := Value;
  except
    // TODO: handle exception
  end;
end;

constructor TPixelGrid.Create(AWidth, AHeight: Integer);
begin
  SetLength(FGrid, AWidth, AHeight);
end;

procedure TPixelGrid.BlurIt(Faktor: Integer);
var
  NewGrid: T2DArrayofInteger;
  X, Y: Integer;
begin
  SetLength(NewGrid, GetWidth, GetHeight);

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      SetPixelValue(X, Y, NewGrid, GetAverage(X, Y, Faktor, Faktor));

  FGrid := NewGrid;
end;

function TPixelGrid.Clean: Boolean;
var
  TopLines, BottomLines, LeftLines, RightLines, Avg, X, Y: Integer;
  RowIsClear, LineIsClear: Boolean;
  Res: T2DArrayofInteger;
begin
  TopLines := 0;
  BottomLines := 0;
  LeftLines := 0;
  RightLines := 0;
  Avg := getAverage;

  for X := 0 to GetWidth - 1 do
  begin
    RowIsClear := True;

    for Y := 0 to GetHeight - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
      begin
        RowIsClear := False;
        Break;
      end;
    end;

    if not RowIsClear then
      Break;

    Inc(LeftLines);
  end;

  for X := GetWidth - 1 downto 0 do
  begin
    RowIsClear := True;

    for Y := 0 to GetHeight - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
      begin
        RowIsClear := False;
        Break;
      end;
    end;

    if not RowIsClear then
      Break;

    Inc(RightLines);
  end;

  if (LeftLines >= GetWidth) or (GetWidth - RightLines > GetWidth) then
  begin
    FGrid := nil;
    SetLength(FGrid, 0, 0);
    Logger.Severe('Cleaning failed. Nothing left (1)');
    Exit(False);
  end;

  for Y := 0 to GetHeight - 1 do
  begin
    LineIsClear := True;
    for X := LeftLines to GetWidth - RightLines - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
      begin
        LineIsClear := False;
        Break;
      end;
    end;

    if not LineIsClear then
      Break;

    Inc(TopLines);
  end;

  for Y := GetHeight - 1 downto 0 do
  begin
    LineIsClear := True;

    for X := LeftLines to GetWidth - RightLines - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
      begin
        LineIsClear := False;
        Break;
      end;
    end;

    if not LineIsClear then
      Break;

    Inc(BottomLines);
  end;

  if (GetWidth - LeftLines - RightLines < 0) or (GetHeight - TopLines - BottomLines < 0) then
  begin
    FGrid := nil;
    SetLength(FGrid, 0, 0);
    Logger.Severe('Cleaning failed. Nothing left! (2)');
    Exit(False);
  end;

  SetLength(Res, GetWidth - LeftLines - RightLines, GetHeight - TopLines - BottomLines);
  FLocation[0] := FLocation[0] + LeftLines;
  FLocation[1] := FLocation[1] + TopLines;

  for Y := 0 to GetHeight - TopLines - BottomLines - 1 do
    for X := 0 to GetWidth - LeftLines - RightLines - 1 do
      Res[X][Y] := GetPixelValue(X + LeftLines, Y + TopLines);

  FGrid := Res;
  Result := True;
end;

procedure TPixelGrid.CleanBackgroundByColor(Avg: Integer);
var
  X, Y, Dif: Integer;
begin
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      Dif := Abs(Avg - GetPixelValue(X, Y));

      if Dif < Trunc(GetMaxPixelValue * (FOwner as TAntiCaptcha).Script.GetParam('BackgroundSampleCleanContrast')) then
        SetPixelValue(X, Y, GetMaxPixelValue);
    end;
end;

procedure TPixelGrid.CleanBackgroundBySample(PX, PY, AWidth, AHeight: Integer);
var
  Avg: Integer;
begin
  Avg := GetAverage(pX + AWidth div 2, PY + AHeight div 2, AWidth, AHeight);
  CleanBackgroundByColor(Avg);
end;

procedure TPixelGrid.ColorObject(PObject: TPixelObject; Color: Integer);
var
  i: Integer;
begin
  for i := 0 to PObject.Size - 1 do
    SetPixelValue(PObject.ElementAt(i)[0], PObject.ElementAt(i)[1], Color);
end;

function TPixelGrid.CreateLetter: TObject;
begin
  Result := TLetter.Create;
  TLetter(Result).Owner := FOwner;
end;

procedure TPixelGrid.Crop(LeftPadding, TopPadding, RightPadding, BottomPadding: Integer);
var
  NewHeight, NewWidth, X, Y: Integer;
  NewGrid: T2DArrayofInteger;
begin
  NewWidth := GetWidth - (LeftPadding + RightPadding);
  NewHeight := GetHeight - (TopPadding + BottomPadding);

  SetLength(NewGrid, NewWidth, NewHeight);
  FLocation[0] := FLocation[0] + LeftPadding;
  FLocation[1] := FLocation[1] + TopPadding;

  for X := 0 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
      NewGrid[X][Y] := FGrid[X + LeftPadding][Y + TopPadding];

  FGrid := NewGrid;
end;

procedure TPixelGrid.DesinX(Max: Double; Omega: Double; Phi: Double);
var
  Tmp: T2DArrayofInteger;
  Shift, Y, X: Integer;
begin
  Omega := 2 * Pi / Omega;

  SetLength(Tmp, GetWidth, GetHeight);

  for Y := 0 to GetHeight - 1 do
  begin
    Shift := Trunc(Max * Sin(Omega * (Y + Phi)));

    for X := 0 to GetWidth - 1 do
    begin
      if (X + Shift < GetWidth) and (X + Shift >= 0) then
        Tmp[X][Y] := FGrid[X + Shift][Y]
      else
        Tmp[X][Y] := $FF;
    end;
  end;

  FGrid := Tmp;
end;

procedure TPixelGrid.DesinY(Max: Double; Omega: Double; Phi: Double);
var
  Tmp: T2DArrayofInteger;
  Shift, Y, X: Integer;
begin
  Omega := 2 * Pi / Omega;

  SetLength(Tmp, GetWidth, GetHeight);

  for X := 0 to GetWidth - 1 do
  begin
    Shift := Trunc(Max * Sin(Omega * (X + Phi)));

    for Y := 0 to GetHeight - 1 do
    begin
      if (Y + Shift < GetHeight) and (Y + Shift >= 0) then
        Tmp[X][Y] := FGrid[X][Y + Shift]
      else
        Tmp[X][Y] := $FF;
    end;
  end;

  FGrid := Tmp;
end;

function TPixelGrid.GetAverage: Integer;
var
  Avg, Bv: TArrayofInteger;
  i, X, Y: Integer;
begin
  i := 0;
  SetLength(Avg, 3);
  Avg[0] := 0;
  Avg[1] := 0;
  Avg[2] := 0;

  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
    begin
      Bv := HexToRGB(GetPixelValue(X, Y));

      Avg[0] := Avg[0] + Bv[0];
      Avg[1] := Avg[1] + Bv[1];
      Avg[2] := Avg[2] + Bv[2];

      Inc(i);
      Bv := nil;
    end;

  if i = 0 then
    Exit(0);

  Avg[0] := Avg[0] div i;
  Avg[1] := Avg[1] div i;
  Avg[2] := Avg[2] div i;
  Result := RGB(Avg[0], Avg[1], Avg[2]);

  Avg := nil;
end;

function TPixelGrid.GetAverage(PX, PY, AWidth, AHeight: Integer): Integer;
var
  Avg, Bv: TArrayofInteger;
  i, X, Y, HalfW, HalfH: Integer;
begin
  i := 0;
  SetLength(Avg, 3);
  Avg[0] := 0;
  Avg[1] := 0;
  Avg[2] := 0;
  HalfW := AWidth div 2;
  HalfH := AHeight div 2;

  if (AWidth = 1) and (PX = 0) then
    AWidth := 2;
  if (AHeight = 1) and (PY = 0) then
    AHeight := 2;

  for X := Max(0, PX - HalfW) to Min(PX + AWidth - HalfW, GetWidth) - 1 do
    for Y := Max(0, PY - HalfH) to Min(PY + AHeight - HalfH, GetHeight) - 1 do
    begin
      Bv := HexToRGB(GetPixelValue(X, Y));
      Avg[0] := Avg[0] + Bv[0];
      Avg[1] := Avg[1] + Bv[1];
      Avg[2] := Avg[2] + Bv[2];

      Inc(i);
      Bv := nil;
    end;

  Avg[0] := Avg[0] div i;
  Avg[1] := Avg[1] div i;
  Avg[2] := Avg[2] div i;
  Result := RGB(Avg[0], Avg[1], Avg[2]);
end;

function TPixelGrid.GetAverageWithoutPoint(PX, PY, AWidth, AHeight: Integer): Integer;
var
  Avg, Bv: TArrayofInteger;
  i, X, Y, HalfW, HalfH: Integer;
begin
  i := 0;
  SetLength(Avg, 3);
  Avg[0] := 0;
  Avg[1] := 0;
  Avg[2] := 0;
  HalfW := AWidth div 2;
  HalfH := AHeight div 2;

  if (AWidth = 1) and (PX = 0) then
    AWidth := 2;
  if (AHeight = 1) and (PY = 0) then
    AHeight := 2;

  for X := Max(0, PX - HalfW) to Min(PX + AWidth - HalfW, GetWidth) - 1 do
    for Y := Max(0, PY - HalfH) to Min(PY + AHeight - HalfH, GetHeight) - 1 do
    begin
      if (X <> PX) or (Y <> PY) then
      begin
        Bv := HexToRGB(GetPixelValue(X, Y));
        Avg[0] := Avg[0] + Bv[0];
        Avg[1] := Avg[1] + Bv[1];
        Avg[2] := Avg[2] + Bv[2];

        Inc(i);
        Bv := nil;
      end;
    end;

  if i > 0 then
  begin
    Avg[0] := Avg[0] div i;
    Avg[1] := Avg[1] div i;
    Avg[2] := Avg[2] div i;
  end;

  Result := RGB(Avg[0], Avg[1], Avg[2]);
end;

function TPixelGrid.GetColorObjects(LetterNum: Integer): TArrayofPixelObject;
(*var
  avg, intensivity, h, d, x, y, key, i: Integer;
  last, rgbA, bv, bestKey: TArrayofInteger;
  map: THashMap;
  found: Boolean;
  bestValue, dif: Double;     *)
begin
 (* Form1.LogMemo.Lines.Add('Max pixel value: ' + IntToStr(Self.GetMaxPixelValue));

  // Erstelle Farbverteilungsmap
  map := THashMap.Create;
  avg := getAverage;
  intensivity := 8;
  h := GetWidth div letterNum div 4;
  d := 0;
  for X := 0 to GetWidth -1 do
    for y := 0 to GetHeight - 1 do
    begin
      key := GetPixelValue(x, y);
      rgbA := hexToRgb(key);

      if (IsElement(key, avg)) or (rgbToHsb(rgbA)[0] * 100 > 0) then
      begin
        if (map.get(TObject(key)) = nil) then
        begin
          Inc(d);
          if (d < GetHeight * 2) then
          begin
            d := 0;
            bv := hexToRgb(key);
            found := False;
            if (last <> nil) and (getHsbColorDifference(bv, hexToRgb(TPixelObject(map.get(TObject(last))).getAverage)) < intensivity) then
            begin
              TPixelObject(map.get(TObject(last))).add(x, y, key);
              found := True;
            end
            else
            begin
              SetLength(bestKey, 2);
              bestKey[0] := -1;
              bestKey[1] := -1;
              bestValue := 255;
              dif := 255;
              //for i := 0 to map.ItemCount - 1 do
              //FEATURE UNIMPLEMENTED !!!
            end;
          end;
        end;
      end;
    end;       *)
end;

function TPixelGrid.GetDim: string;
begin
  Result := Format('(%d/%d)', [GetWidth, GetHeight]);
end;

function TPixelGrid.GetFullImage: TBitmap;
var
  X, Y: Integer;
begin
  if (GetWidth <= 0) or (GetHeight <= 0) then
    raise Exception.Create('Dimension falsch!');

  Result := TBitmap.Create;
  Result.Width := GetWidth;
  Result.Height := GetHeight;
  Result.PixelFormat := pf8bit;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
      if TColor(GetPixelValue(X, Y)) = 0 then
        Result.Canvas.Pixels[X, Y] := 0
      else
        Result.Canvas.Pixels[X, Y] := $FFFFFF;
end;

function TPixelGrid.GetGridCopy: T2DArrayofInteger;
var
  X, Y: Integer;
begin
  SetLength(Result, GetWidth, GetHeight);
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      Result[X][Y] := FGrid[X][Y];
end;

function TPixelGrid.GetHeight: Integer;
begin
  if Length(FGrid) < 1 then
    Exit(0);

  Result := Length(FGrid[0]);
end;

function TPixelGrid.GetImage: TBitmap;
var
  X, Y: Integer;
begin
  if (GetWidth <= 0) or (GetHeight <= 0) then
    raise Exception.Create('Dimension falsch!');

  Result := TBitmap.Create;
  Result.Width := GetWidth;
  Result.Height := GetHeight;
  Result.PixelFormat := pf8bit;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
      Result.Canvas.Pixels[X, Y] := TColor(GetPixelValue(X, Y));
end;

function TPixelGrid.GetImage(Faktor: Integer): TBitmap;
var
  X, Y: Integer;
begin
  if ((GetWidth * Faktor) <= 0) or ((GetHeight * Faktor) <= 0) then
    raise Exception.Create('Bild zu klein! Buchstaben nicht richtig erkannt?');

  Result := TBitmap.Create;
  Result.Width := GetWidth * Faktor;
  Result.Height := GetHeight * Faktor;
  Result.PixelFormat := pf8bit;

  Y := 0;
  X := 0;

  while Y < (GetHeight * Faktor) do
  begin
    while X < (GetWidth * Faktor) do
    begin
      Result.Canvas.Brush.Color := TColor(GetPixelValue(X div Faktor, Y div Faktor));
      Result.Canvas.FillRect(Rect(X, Y, Faktor, Faktor));
      X := X + Faktor;
    end;
    Y := Y + Faktor;
  end;
end;

function TPixelGrid.GetMaxPixelValue: Integer;
begin
  Result := GetMaxPixelValue(FOwner);
end;

function TPixelGrid.GetMaxPixelValue(Faktor: Double): Integer;
begin
  Result := Trunc(GetMaxPixelValue(FOwner) * Faktor);
end;

(* Diese rekusrive Funktion kann bei zu groen Objekten zu einem Stackoverflow fhren. Man sollte sie mal umschreiben! *)
procedure TPixelGrid.GetObject(X: Integer; Y: Integer; TmpGrid: T2DArrayofInteger; PObject: TPixelObject);
var
  LocalValue: Integer;
begin
  if (X < 0) or (Y < 0) or (Length(TmpGrid) <= X) or (Length(TmpGrid[0]) <= Y) or (TmpGrid[X][Y] < 0) then
    Exit;

  LocalValue := GetPixelValue(X, Y, TmpGrid);

  if PObject.DoesColorAverageFit(LocalValue) then
  begin
    PObject.Add(X, Y, LocalValue);
    TmpGrid[X][Y] := -1;

    // Achtung!! Algos funktionieren nur auf sw basis richtig
    // grid[X][Y] = 254;
    GetObject(X - 1, Y, TmpGrid, PObject);

    if Boolean((FOwner as TAntiCaptcha).Script.GetParam('followXLines')) then
      GetObject(X - 1, Y - 1, TmpGrid, PObject);

    GetObject(X, Y - 1, TmpGrid, PObject);

    if Boolean((FOwner as TAntiCaptcha).Script.GetParam('followXLines')) then
      GetObject(X + 1, Y - 1, TmpGrid, PObject);

    GetObject(X + 1, Y, TmpGrid, PObject);

    if Boolean((FOwner as TAntiCaptcha).Script.GetParam('followXLines')) then
      GetObject(X + 1, Y + 1, TmpGrid, PObject);

    GetObject(X, Y + 1, TmpGrid, PObject);

    if Boolean((FOwner as TAntiCaptcha).Script.GetParam('followXLines')) then
      GetObject(X - 1, Y + 1, TmpGrid, PObject);
  end;
end;

function TPixelGrid.GetObjects(Contrast, ObjectContrast: Double): TList<TPixelObject>;
var
  TmpGrid: T2DArrayofInteger;
  Dist, X, Y, i, Tmp: Integer;
  Res: TList<TPixelObject>;
  LastObject, PObject: TPixelObject;
begin
  TmpGrid := GetGridCopy;

  Res := TList<TPixelObject>.Create(TComparer<TPixelObject>.Construct(PixelObjectComparer));

  LastObject := nil;

  {$IFDEF DEBUG}
  frmGUI.ShowImage(Self.GetImage);
  {$ENDIF}

  for X := 0 to GetWidth -1 do
    for Y := 0 to GetHeight - 1 do
    begin
      if (TmpGrid[X][Y] < 0) then
        Continue;

      if GetPixelValue(X, Y) <= (ObjectContrast * GetMaxPixelValue) then
      begin
        // Fge 2 Objekte zusammen die scheinbar zusammen gehren
        Dist := 100;
        if Assigned(LastObject) then
          Dist := Trunc((Power(X - (LastObject.XMin + LastObject.Width / 2), 2) + Power(Y - (LastObject.YMin + LastObject.Height / 2), 2)));

        if (Assigned(LastObject)) and
           (LastObject.Area < Integer((FOwner as TAntiCaptcha).Script.GetParam('minimumObjectArea'))) and
           (Dist < Power(Integer((FOwner as TAntiCaptcha).Script.GetParam('minimumLetterWidth')) / 2 + 1, 2)) then
        begin
          PObject := LastObject;
          for i := 0 to Res.Count - 1 do
            if Res[i] = PObject then
            begin
              Res.Delete(i);
              Break;
            end;

          Logger.Finer('Verfolge weiter Letztes Object: Area:' + IntToStr(LastObject.Area) + ' Dist: ' + IntToStr(dist));
        end
        else
        begin
          PObject := TPixelObject.Create(Self);
          PObject.Contrast := Contrast;
          PObject.WhiteContrast := ObjectContrast;
        end;

        {$IFDEF DEBUG}
        if PObject.Area > 20 then
        begin
          frmGUI.ShowImage(getImage, 'Area: ' + IntToStr(PObject.Area));
          frmGUI.ShowImage(TLetter(pObject.ToLetter).GetImage, TLetter(PObject.toLetter).getDim);
        end;
        {$ENDIF}

        Tmp := PObject.Size;
        GetObject(X, Y, TmpGrid, PObject);

        if Tmp = PObject.Size then
        begin
          PObject := TPixelObject.Create(Self);
          PObject.Contrast := Contrast;
          PObject.WhiteContrast := objectContrast;
          GetObject(X, Y, TmpGrid, PObject);
        end;

        {$IFDEF DEBUG}
        if PObject.Area > 20 then
          frmGUI.ShowImage(TLetter(PObject.ToLetter).GetImage, TLetter(PObject.ToLetter).GetDim);
        {$ENDIF}

        LastObject := PObject;

        for i := 0 to Res.Count - 1 do
          if PObject.Area > TPixelObject(Res[i]).Area then
          begin
            Res.Insert(i, PObject);
            PObject := nil;

            Break;
          end;

        if Assigned(PObject) then
          Res.Add(PObject);
      end;
    end;

  Result := Res;
end;

function TPixelGrid.GetPixel: TArrayofInteger;
var
  Pix: TArrayofInteger;
  Pixel, X, Y: Integer;
begin
  SetLength(Pix, GetWidth * GetHeight);
  Pixel := 0;

  for Y := 0 to GetHeight -1 do
    for X := 0 to GetWidth - 1 do
    begin
      Pix[Pixel] := GetPixelValue(X, Y);
      Inc(Pixel);
    end;

  Result := Pix;
end;

function TPixelGrid.GetPixelValue(X, Y: Integer): Integer;
begin
  Result := GetPixelValue(X, Y, FGrid);
end;

function TPixelGrid.GetString: string;
var
  Avg, X, Y: Integer;
begin
  Avg := GetAverage;

  for Y := 0 to GetHeight -1 do
  begin
    for X := 0 to GetWidth - 1 do
    begin
      if IsElement(GetPixelValue(X, Y), Avg) then
        Result := Result + '*'
      else
        Result := Result + IntToStr(Floor(9 * (GetPixelValue(X, Y) / GetMaxPixelValue)));
    end;

    Result := Result + sLineBreak;
  end;
end;

function TPixelGrid.GetWidth: Integer;
begin
  Result := Length(FGrid);
end;

procedure TPixelGrid.Invert;
var
  X, Y: Integer;
  A, B: TArrayofInteger;
begin
  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      A := hexToRgb(GetMaxPixelValue);
      B := hexToRgb(GetPixelValue(X, Y));
      A[0] := A[0] - B[0];
      A[1] := A[1] - B[1];
      A[2] := A[2] - B[2];
      SetPixelValue(X, Y, RGB(A[0], A[1], A[2]));
    end;
end;

function TPixelGrid.IsElement(Value, Avg: Integer): Boolean;
var
  RelativeContrast: Double;
begin
  try
    RelativeContrast := Double((FOwner as TAntiCaptcha).Script.GetParam('relativeContrast'));
  except
    RelativeContrast := -1;
  end;

  if RelativeContrast = -1 then
    Logger.Warning('Kein Double-Parameter fr "RelativeContrast"!');

  if (FOwner as TAntiCaptcha).Script.GetParam('colorType') = 'b' then
    Value := GetBValue(Value);

  Result := Value < Avg * RelativeContrast;
end;

procedure TPixelGrid.Normalize;
begin
  Normalize(1);
end;

procedure TPixelGrid.Normalize(Multi: Double);
begin
  Normalize(Multi, 0, 0);
end;

procedure TPixelGrid.Normalize(Multi: Double; CutMax: Double; CutMin: Double);
var
  Max, Min, Akt, X, Y, Faktor: Integer;
begin
  Max := 0;
  Min := MAXINT;
  CutMin := CutMin * GetMaxPixelValue;
  CutMax := CutMax * GetMaxPixelValue;
  CutMax := GetMaxPixelValue - CutMax;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      Akt := GetPixelValue(X, Y);
      if (Akt < Min) and (Akt > cutMin) then
        Min := Akt;
      if (Akt > Max) and (Akt < cutMax) then
        Max := Akt;
    end;

  Faktor := (Max - Min) div GetMaxPixelValue;
  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      Akt := GetPixelValue(X, Y);
      if Akt <= CutMin then
      begin
        SetPixelValue(X, Y, 0);
        Continue;
      end;

      if Akt >= CutMax then
      begin
        SetPixelValue(X, Y, GetMaxPixelValue);
        Continue;
      end;

      Akt := Akt - Min;
      Akt := Akt div Faktor;
      Akt := Trunc(Akt * multi);
      Akt := Math.Min(Akt, GetMaxPixelValue);
      Akt := Math.Max(Akt, 0);
      SetPixelValue(X, Y, Akt);
    end;
end;

procedure TPixelGrid.PrintGrid;
begin
  Logger.Info(sLineBreak + GetString);
end;

procedure TPixelGrid.ReduceBlackNoise(Faktor: Integer);
begin
  ReduceBlackNoise(Faktor, 1.0);
end;

procedure TPixelGrid.ReduceBlackNoise(Faktor: Integer; Contrast: Double);
var
  Avg, X, Y, LocalAVG: Integer;
  NewGrid: T2DArrayofInteger;
begin
  Avg := GetAverage;
  SetLength(NewGrid, GetWidth, GetHeight);

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
      // Korrektur weil sonst das linke obere PUxel schwarz wird.
      if (X = 0) and (Y = 0) and (Faktor < 3) then
        NewGrid[0][0] := FGrid[0][0]
      else
      begin
        LocalAVG := GetAverageWithoutPoint(X, Y, Faktor, Faktor);

        if (IsElement(GetPixelValue(X, Y), Trunc(Avg * contrast))) and (LocalAVG >= (Contrast * GetMaxPixelValue)) then
          SetPixelValue(X, Y, NewGrid, LocalAVG)
        else
          SetPixelValue(X, Y, NewGrid, GetPixelValue(X, Y));
      end;

  FGrid := NewGrid;
end;

procedure TPixelGrid.ReduceWhiteNoise(Faktor: Integer);
begin
  ReduceWhiteNoise(Faktor, 1.0);
end;

procedure TPixelGrid.ReduceWhiteNoise(Faktor: Integer; Contrast: Double);
var
  Avg, X, Y: Integer;
  NewGrid: T2DArrayofInteger;
begin
  Avg := GetAverage;
  SetLength(NewGrid, GetWidth, GetHeight);

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
      // Korrektur weil sonst das linke obere PUxel schwarz wird.
      if (X = 0) and (Y = 0) and (faktor < 3) then
        NewGrid[0][0] := FGrid[0][0]
      else
        if (not IsElement(GetPixelValue(X, Y), Trunc(Avg * Contrast))) then
          SetPixelValue(X, Y, NewGrid, GetAverageWithoutPoint(X, Y, Faktor, Faktor));

  FGrid := NewGrid;
end;

procedure TPixelGrid.RemoveBridges(Pixels: Integer; Middle: Double);
var
  Avg, Ignorh1, Ignorh2, X, Y, C, i: Integer;
  NewGrid: T2DArrayofInteger;
begin
  Avg := getAverage;
  SetLength(NewGrid, GetWidth, GetHeight);
  Ignorh2 := Trunc(GetHeight / Middle);
  Ignorh1 := GetHeight - Ignorh2;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      if (X < Pixels) or (Y < Pixels) or (Y > Ignorh1) or (Y < Ignorh2) then
        NewGrid[X][Y] := FGrid[X][Y]
      else
      begin
        if IsElement(GetPixelValue(X, Y), Avg) then
        begin
          C := 0;
          i := 0;
          while (C <= pixels) and (Y + i < GetHeight) do
          begin
            if (IsElement(GetPixelValue(X, Y + i), Avg)) or (IsElement(GetPixelValue(X + 1, Y + i), Avg)) or (IsElement(GetPixelValue(X - 1, Y + i), Avg)) or (IsElement(GetPixelValue(X + 1, Y + i - 1), Avg)) or (IsElement(GetPixelValue(X - 1, Y + i - 1), Avg)) then
              Inc(C)
            else
              Break;

            Inc(i);
          end;
          i := 0;
          while (C <= pixels) and (Y - i > 0) do
          begin
            if (IsElement(GetPixelValue(X, Y - i), Avg)) or (IsElement(GetPixelValue(X + 1, Y - i), Avg)) or (IsElement(GetPixelValue(X - 1, Y - i), Avg)) or (IsElement(GetPixelValue(X + 1, Y - i - 1), Avg)) or (IsElement(GetPixelValue(X - 1, Y - i - 1), Avg)) then
              Inc(C)
            else
              Break;

            Inc(i);
          end;

          if C <= Pixels then
            SetPixelValue(X, Y, NewGrid, GetMaxPixelValue)
          else
            NewGrid[X][Y] := FGrid[X][Y];
        end
        else
          NewGrid[X][Y] := FGrid[X][Y];
      end;
    end;

  FGrid := NewGrid;
end;

procedure TPixelGrid.RemoveObjectFromGrid(PObject: TPixelObject);
begin
  ColorObject(PObject, GetMaxPixelValue);
end;

procedure TPixelGrid.RemoveSmallObjects(Contrast: Double; ObjectContrast: Double);
var
  Tmp, i: Integer;
  Res: TList<TPixelObject>;
begin
  try
    Tmp := Integer((FOwner as TAntiCaptcha).Script.GetParam('minimumObjectArea'));
  except
    raise Exception.Create('Kein Integer-Parameter fr "minimumObjectArea"!');
  end;

  (FOwner as TAntiCaptcha).Script.SetParam('minimumObjectArea', 0);
  Res := GetObjects(Contrast, ObjectContrast);
  (FOwner as TAntiCaptcha).Script.SetParam('minimumObjectArea', Tmp);

  for i := 1 to Res.Count - 1 do
    RemoveObjectFromGrid(Res[i]);
end;

procedure TPixelGrid.RemoveSmallObjects(Contrast, ObjectContrast: Double; MaxSize: Integer);
var
  Tmp, i: Integer;
  Res: TList<TPixelObject>;
begin
  try
    Tmp := Integer((FOwner as TAntiCaptcha).Script.GetParam('minimumObjectArea'));
  except
    raise Exception.Create('Kein Integer-Parameter fr "minimumObjectArea"!');
  end;

  (FOwner as TAntiCaptcha).Script.SetParam('minimumObjectArea', 0);
  Res := GetObjects(Contrast, ObjectContrast);
  (FOwner as TAntiCaptcha).Script.SetParam('minimumObjectArea', Tmp);

  for i := 1 to Res.Count - 1 do
    if TPixelObject(Res[i]).Size < MaxSize then
      RemoveObjectFromGrid(Res[i]);
end;

procedure TPixelGrid.SampleDown(Faktor: Integer);
var
  NewHeight, NewWidth, X, Y, LocalAVG, Values, GX, GY, NewX, NewY: Integer;
  NewGrid: T2DArrayofInteger;
begin
  NewWidth := Math.Ceil(GetWidth / Faktor);
  NewHeight := Math.Ceil(GetHeight / Faktor);

  SetLength(NewGrid, GetWidth, GetHeight);
  for X := 0 to NewWidth - 1 do
    for Y := 0 to NewHeight - 1 do
    begin
      LocalAVG := 0;
      Values := 0;

      for GX := 0 to Faktor - 1 do
        for GY := 0 to Faktor - 1 do
        begin
          NewX := X * Faktor + GX;
          NewY := Y * Faktor + GY;

          if (NewX > GetWidth) or (NewY > GetHeight) then
            Continue;

          LocalAVG := MixColors(LocalAVG, GetPixelValue(NewX, NewY), Values, 1);
          Inc(Values);
        end;

      for GX := 0 to Faktor - 1 do
        for GY := 0 to Faktor - 1 do
        begin
          NewX := X * Faktor + GX;
          NewY := Y * Faktor + GY;
          SetPixelValue(NewX, NewY, NewGrid, LocalAVG);
        end;
    end;

  FGrid := NewGrid;
end;

procedure TPixelGrid.SaveImageAsJpeg(FileName: string);
var
  Img: TBitmap;
  JPG: TJPEGImage;
  X, Y: Integer;
begin
  Img := TBitmap.Create;
  Img.SetSize(GetWidth, GetHeight);
  Img.PixelFormat := pf24bit;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
      Img.Canvas.Pixels[X, Y] := GetPixelValue(X, Y);

  JPG := TJPEGImage.Create;
  try
    JPG.PixelFormat := jf24Bit;
    JPG.Assign(Img);
    JPG.SaveToFile(FileName);
  finally
    JPG.Free;
    Img.Free;
  end;
end;

//    public void setGridCopy(int[][] grid, int leftPadding, int topPadding, int rightPadding, int bottomPadding)
//    FEATURE UNIMPLEMENTED !!!

procedure TPixelGrid.SetOrgGrid(AGrid: T2DArrayofInteger);
begin
  FTmpGrid := AGrid;
end;

procedure TPixelGrid.SetPixel(APixel: TArrayofInteger);
var
  i, X, Y: Integer;
begin
  FPixel := APixel;
  i := 0;

  for Y := 0 to GetHeight - 1 do
    for X := 0 to GetWidth - 1 do
    begin
      Inc(i);
      FGrid[X][Y] := APixel[i];
    end;
end;

procedure TPixelGrid.SetPixelValue(X, Y, Value: Integer);
begin
  SetPixelValue(X, Y, FGrid, Value);
end;

procedure TPixelGrid.TestColor;
var
  X, Y: Integer;
begin
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      // Einmal um die Farbe und wieder zurck
      SetPixelValue(X, Y, GetPixelValue(X, Y));
end;

procedure TPixelGrid.ToBlackAndWhite;
begin
  ToBlackAndWhite(1);
end;

procedure TPixelGrid.ToBlackAndWhite(Contrast: Double);
var
  X, Y: Integer;
begin
  for X := 0 to GetWidth - 1 do
    for Y := 0 to GetHeight - 1 do
      if IsElement(GetPixelValue(X, Y), Trunc(GetMaxPixelValue * Contrast)) then
        SetPixelValue(X, Y, 0)
      else
        SetPixelValue(X, Y, GetMaxPixelValue);
end;

function TPixelGrid.ToHsbColorString: string;
var
  X, Y: Integer;
  RGB: TArrayofInteger;
  HSB: TArrayofDouble;
begin
  Result := '';

  for X := 0 to GetWidth -1 do
    for Y := 0 to GetHeight - 1 do
    begin
      RGB := HexToRgb(GetPixelValue(X, Y));
      HSB := RgbToHsb(RGB);
      Result := Result + Format('y(%d)x(%d)=%f', [Y, X, HSB[0] * 100]) + sLineBreak;
    end;
end;

end.
