(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit EasyCaptcha;

interface

uses Captcha, Variants, Generics.Collections, CPoint, XMLDoc, XMLIntf, SysUtils,
     Utilities, PixelObject, Math, PixelGrid, Generics.Defaults, Letter;

type
  TArrayofVariant = array of Variant;

  TEasyCaptcha = class
  public
    class function MergeObjectsBasic(var OS: TList<TPixelObject>; Captcha: TCaptcha; Gap: Integer): TArrayofInteger;
    class function Clean(var Captcha: TCaptcha): TArrayofVariant;
    class function GetLetters(Captcha: TCaptcha): TArrayofLetter;
  end;

implementation

uses AntiCaptcha, untLogger;

{ TEasyCaptcha }

function LoadCPointsFromXML(FileName: string): TList<TCPoint>;
var
  XML: IXMLDocument;
  Nodes, Props: IXMLNodeList;
  i, j: Integer;
  Prop: string;
  CurPoint: TCPoint;
begin
  XML := NewXMLDocument;
  XML.LoadFromFile(FileName);
  XML.Active := True;
  Nodes := XML.DocumentElement.ChildNodes[0].ChildNodes;

  Result := TList<TCPoint>.Create;

  // Iterate through "add" voids
  for i := 0 to Nodes.Count - 1 do
  begin
    Props := Nodes[i].ChildNodes[0].ChildNodes;
    for j := 0 to Props.Count - 1 do
    begin
      Prop := Props[j].Attributes['property'];

      // Std values
      CurPoint.ColorDifferenceMode := LAB_DIFFERENCE;
      CurPoint.Foreground := True;

      if Prop = 'color' then
        CurPoint.Color := StrToInt(Props[j].ChildNodes[0].Text)
      else
      if Prop = 'colorDistanceMode' then
        CurPoint.ColorDifferenceMode := StrToInt(Props[j].ChildNodes[0].Text)
      else
      if Prop = 'distance' then
        CurPoint.Distance := StrToInt(Props[j].ChildNodes[0].Text)
      else
      if Prop = 'foreground' then
        CurPoint.Foreground := StrToBool(Props[j].ChildNodes[0].Text);
    end;

    Result.Add(CurPoint);
  end;
end;

function GetSWCaptcha(Obj: TPixelObject): TCaptcha;
var
  LGrid: TGrid;
  X, Y, d: Integer;
  Akt: TArrayofInteger;
begin
  SetLength(LGrid, TPixelGrid(Obj.Owner).Width, TPixelGrid(Obj.Owner).Height);
  for X := 0 to High(LGrid) do
    for Y := 0 to TPixelGrid(Obj.Owner).Height - 1 do
      LGrid[X][Y] := $FFFFFF;

  for d := 0 to Obj.Size - 1 do
  begin
    Akt := Obj.ElementAt(d);
    LGrid[Akt[0]][Akt[1]] := 0;
  end;

  Result := TCaptcha.Create(TPixelGrid(Obj.Owner).Width, TPixelGrid(Obj.Owner).Height);
  Result.Owner := TPixelGrid(Obj.Owner).Owner;
  Result.Grid := LGrid;
end;

procedure MergeOS(AOS: TPixelObject; var OS: TList<TPixelObject>);
var
  NextOS, b: TPixelObject;
  Best, Matching, i, ib: Integer;
begin
  OS.Remove(AOS);
  NextOS := OS[0];
  Matching := Min(AOS.XMax, NextOS.XMax) - Max(AOS.XMin, NextOS.XMin);
  if Matching >= 0 then
    Best := -Matching
  else
    Best := Min(Min(Abs((AOS.XMin + AOS.Width) - (NextOS.XMin)),
                    Abs(AOS.XMin - (NextOS.XMin + NextOS.Width))),
                Min(Abs(AOS.XMin - NextOS.XMin),
                    Abs((AOS.XMin + AOS.Width) - (NextOS.XMin + NextOS.Width))));

  for i := 1 to OS.Count - 1 do
  begin
    b := OS[i];
    Matching := Min(AOS.XMax, b.XMax) - Max(AOS.XMin, b.XMin);
    if Matching >= 0 then
      ib := -Matching
    else
      ib := Min(Min(Abs((AOS.XMin + AOS.Width) - (b.XMin)),
                    Abs(AOS.XMin - (b.XMin + b.Width))),
                Min(Abs(AOS.XMin - b.XMin),
                    Abs((AOS.XMin + AOS.Width) - (b.XMin + b.Width))));

    if ib < Best then
    begin
      Best := ib;
      NextOS := b;
    end;
  end;
  NextOS.Add(AOS);
end;

class function TEasyCaptcha.MergeObjectsBasic(var OS: TList<TPixelObject>;
  Captcha: TCaptcha; Gap: Integer): TArrayofInteger;
var
  Area, OSCount, i, j, Best, ib, Matching: Integer;
  AOS, Biggest, PO, PO2, NextOS: TPixelObject;
begin
  Area := 0;
  Result := nil;
  OSCount := 0;

  try
    AOS := OS[0];
    Biggest := AOS;
    for PO in OS do
    begin
      Inc(Area, PO.Area);
      Inc(OSCount, PO.Size);

      if AOS.Width > PO.Width then
        AOS := PO;
      if Biggest.Area < PO.Area then
        Biggest := PO;
    end;
    SetLength(Result, 2);
    Result[0] := Area div (OS.Count * 3);
    Result[1] := OSCount div (OS.Count * 3);

    for PO in OS do
      if (PO.Area > Area / 3) and ((OSCount / TAntiCaptcha(Captcha.Owner).LetterNum) / 5 > PO.Size) then
      begin
        OS.Remove(PO);
        Exit(MergeObjectsBasic(OS, Captcha, Gap));
      end;

    for i := 0 to OS.Count - 1 do
    begin
      PO := OS[i];
      for j := 0 to i - 1 do
      begin
        PO2 := OS[j];
        if PO.Width < PO2.Width then
          if ((PO2.XMin - PO.XMin) <= 0) and (PO2.XMax - PO.XMax >= 0) then
          begin
            PO2.Add(PO);
            OS.Remove(PO);
            Exit(MergeObjectsBasic(OS, Captcha, Gap));
          end;
      end;
    end;

    for i := 0 to OS.Count - 1 do
    begin
      PO := OS[i];
      if PO.Size < OSCount / (TAntiCaptcha(Captcha.Owner).LetterNum / 2) then
      begin
        NextOS := nil;
        Best := MAXINT;
        for j := 0 to i - 1 do
        begin
          PO2 := OS[j];
          if PO2 = PO then
            Continue;

          Matching := Min(AOS.XMax, PO2.XMax) - Max(AOS.XMin, PO2.XMin);
          if Matching >= 0 then
            ib := -Matching
          else
            ib := Min(Min(Abs((AOS.XMin + AOS.Width) - (PO2.XMin)),
                          Abs(AOS.XMin - (PO2.XMin + PO2.Width))),
                      Min(Abs(AOS.XMin - PO2.XMin),
                          Abs((AOS.XMin + AOS.Width) - (PO2.XMin + PO2.Width))));

          if ib < Best then
          begin
            Best := ib;
            NextOS := PO2;
          end;
        end;

        if (Best < Gap / 60) and (NextOS <> PO) then
        begin
          NextOS.Add(PO);
          OS.Remove(PO);
          Exit(MergeObjectsBasic(OS, Captcha, Gap));
        end;
      end;
    end;

    if OS.Count > TAntiCaptcha(Captcha.Owner).LetterNum then
      for i := 0 to OS.Count - 1 do
      begin
        PO := OS[i];
        if PO.Area < Area then
        begin
          NextOS := nil;
          Best := MAXINT;
          for j := 0 to i - 1 do
          begin
            PO2 := OS[j];
            if PO2 = PO then
              Continue;

            Matching := Min(AOS.XMax, PO2.XMax) - Max(AOS.XMin, PO2.XMin);
            if Matching >= 0 then
              ib := -Matching
            else
              ib := Min(Min(Abs((AOS.XMin + AOS.Width) - (PO2.XMin)),
                            Abs(AOS.XMin - (PO2.XMin + PO2.Width))),
                        Min(Abs(AOS.XMin - PO2.XMin),
                            Abs((AOS.XMin + AOS.Width) - (PO2.XMin + PO2.Width))));

            if ib < Best then
            begin
              Best := ib;
              NextOS := PO2;
            end;
          end;

          if (Best < Gap / 60) and (NextOS <> PO) then
          begin
            NextOS.Add(PO);
            OS.Remove(PO);
            Exit(MergeObjectsBasic(OS, Captcha, Gap));
          end;
        end;
      end;

    if (Biggest.Area / (TAntiCaptcha(Captcha.Owner).LetterNum * 2) > AOS.Area) or
       (OSCount / (TAntiCaptcha(Captcha.Owner).LetterNum * 2) > AOS.Size) then
    begin
      MergeOS(AOS, OS);
      Exit(MergeObjectsBasic(OS, Captcha, Gap));
    end;

    if OS.Count <= TAntiCaptcha(Captcha.Owner).LetterNum then
    begin
      Area := Area div (OS.Count * 2);
      OSCount := OSCount div (OS.Count * 2);
    end
    else
    begin
      OSCount := OSCount div OS.Count;
      Area := Area div OS.Count;
    end;

    if (AOS.Area < Area) or (AOS.Size < OSCount) then
    begin
      MergeOS(AOS, OS);
      Exit(MergeObjectsBasic(OS, Captcha, Gap));
    end;

    for PO in OS do
      if (PO <> AOS) and (PO.Size < (AOS.Size / 2)) then
      begin
        MergeOS(PO, OS);
        Exit(MergeObjectsBasic(OS, Captcha, Gap));
      end;
  except
    Logger.Warning('Exception in MergeObjectsBasic: ' + Exception(EXceptObject).Message);
  end;
end;

function GetRightLetters(var OS: TList<TPixelObject>; Captcha: TCaptcha; X: Integer;
  MergeInfos: TArrayofInteger; var i: Integer): TList<TPixelObject>;
var
  PO: TPixelObject;
begin
  Inc(i);

  Logger.Info('StartCount: ' + IntToStr(OS.Count));

  if i > TAntiCaptcha(Captcha.Owner).LetterNum * 2 then
    Exit(OS);

  for PO in OS do
  begin
    if OS.Count < 3 then
      Break;

    if (PO.Area < MergeInfos[0] / (TAntiCaptcha(Captcha.Owner).LetterNum * 4)) or
       (PO.Size < MergeInfos[1] / (TAntiCaptcha(Captcha.Owner).LetterNum * 5)) then
      OS.Remove(PO);
  end;

  if OS.Count >= TAntiCaptcha(Captcha.Owner).LetterNum then
    Exit(OS);

  Logger.Info('EndCount: ' + IntToStr(OS.Count));


  Result := OS;
end;

class function TEasyCaptcha.Clean(var Captcha: TCaptcha): TArrayofVariant;
var
  Points: TList<TCPoint>;
  ResYMax, ResX, Gap, LastGap, NrX, NrY, X, Y, Bcuy, GC, MC: Integer;
  Grid: TGrid;
  BW, Add, Set1, Set2: Boolean;
  ResO, ResO2, Merge, CO: TList<TPixelObject>;
  BestDist1, BestDist2, Dist, RGBDiff: Double;
  PBestDist1, PBestDist2, CP: TCPoint;
  n, o, PO: TPixelObject;
begin
  TAntiCaptcha(Captcha.Owner).Script.ExecutePrepareCommands(Captcha);
  Points := LoadCPointsFromXML(TAntiCaptcha(Captcha.Owner).MethodPath + 'CPoints.xml');

  ResYMax := 0;   // Hhe des grten Letters
  ResX := 0;     // Weite aller Letters
  Gap := 0;
  LastGap := -1;
  Grid := Captcha.Grid;
  BW := TAntiCaptcha(Captcha.Owner).Script.GetParam('easyCaptchaBW');
  ResO := TList<TPixelObject>.Create;
  NrX := Trunc(Captcha.Width / (TAntiCaptcha(Captcha.Owner).LetterNum * 1.2));
  NrY := Captcha.Height;

  // Farbunterscheidung durch Ebenen einbauen
  for X := 0 to Captcha.Width - 1 do
  begin
    Bcuy := 0;
    for Y := 0 to Captcha.Height - 1 do
    begin
      BestDist1 := MaxDouble;
      BestDist2 := MaxDouble;
      Set1 := False;
      Set2 := False;

      for CP in Points do
      begin
        Dist := CP.GetColorDifference(Grid[X][Y]);

        if BestDist1 > Dist then
        begin
          BestDist1 := Dist;
          PBestDist1 := CP;
          Set1 := True;
        end;

        if (Dist < CP.Distance) and (BestDist2 > Dist) then
        begin
          BestDist2 := 0;
          PBestDist2 := CP;
          Set2 := True;
        end;
      end;

      Add := False;
      if Set2 then
      begin
        if not PBestDist2.Foreground then
          Captcha.SetPixelValue(X, Y, $FFFFFF)
        else
        begin
          Add := True;
          if BW then
            Captcha.SetPixelValue(X, Y, 0);
          Inc(BCuy);
        end;
      end
      else
      if Set1 then
      begin
        Captcha.SetPixelValue(X, Y, $FFFFFF);

        if PBestDist1.Foreground then
        begin
          Add := True;
          Inc(BCuy);
        end;
      end
      else
        Add := True;

      if Add then
      begin
        n := TPixelObject.Create(Captcha);
        GC := Captcha.Grid[X][Y];
        n.Add(X, Y, GC);

        Merge := TList<TPixelObject>.Create;
        for o in ResO do
        begin
          MC := o.GetMostColor;
          RGBDiff := GetRGBColorDifference2(GC, MC);
          if (o.IsTouching(X, Y, True, NrX, NrY)) and ((RGBDiff < 6) or ((RGBDiff < 25) and (GetColorDifference(GC, MC) < 8))) then
            Merge.Add(o);
        end;

        case Merge.Count of
          0: ResO.Add(n);
          1: Merge[0].Add(n);
          else
          begin
            for PO in Merge do
            begin
              ResO.Remove(PO);
              n.Add(PO);
            end;
            ResO.Add(n);
          end;
        end;

        Captcha.SetPixelValue(X, Y, 0);
      end;
    end;

    if Bcuy > ResYMax then
      ResYMax := Bcuy;
    if Bcuy > 0 then
    begin
      if LastGap = -1 then
        LastGap := 0
      else
        Inc(LastGap, Gap);
      Inc(ResX);
      Gap := 0;
    end
    else
      Inc(Gap);
  end;

  if BW then
    Captcha.SetOrgGrid(Captcha.Grid);

  Gap := LastGap div TAntiCaptcha(Captcha.Owner).LetterNum;
  ResO2 := TList<TPixelObject>.Create;

  for PO in ResO do
    try
      CO := GetSWCaptcha(PO).GetObjects(0.5, 0.5);

    //  MergeObjectsBasic(CO, Captcha, Gap);   xxx

      for O in CO do
        ResO2.Add(O);
      CO.Free;
    except
    end;

  SetLength(Result, 4);
  Result[0] := ResX;
  Result[1] := ResYMax;
  Result[2] := LastGap;
  Result[3] := Cardinal(ResO2);
end;

class function TEasyCaptcha.GetLetters(Captcha: TCaptcha): TArrayofLetter;
var
  Cl: TArrayofVariant;
  OS: TList<TPixelObject>;
  i: Integer;
 // Gap, Counter: Integer;
  //MergeInfos: TArrayofInteger;
begin
  Cl := Clean(Captcha);
  OS := TList<TPixelObject>(Pointer(Cardinal(Cl[3])));
  try
    OS.Sort(TComparer<TPixelObject>.Construct(PixelObjectComparer));
    (*Gap := Integer(Cl[2]) div TAntiCaptcha(Captcha.Owner).LetterNum;

    xxx   Macht Probleme mit der Abspaltung - Letter verschwinden

    MergeInfos := MergeObjectsBasic(OS, Captcha, Gap);
    Counter := 0;
    GetRightLetters(OS, Captcha, Cl[0], MergeInfos, Counter);
    OS.Sort(TComparer<TPixelObject>.Construct(PixelObjectComparer));   *)

    SetLength(Result, OS.Count);
    for i := 0 to OS.Count - 1 do
    begin
      Result[i] := TLetter(TPixelObject(OS[i]).ToLetter);
      TAntiCaptcha(Captcha.Owner).Script.ExecuteLetterPrepareCommands(Result[i]);
    end;
  finally
    OS.Free;
  end;
end;

end.
