(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit netloadin;

interface

uses AntiCaptcha, Captcha, Letter, Utilities, PixelGrid, EasyCaptcha;

type
  TNtLdN = class(TSpecial)
  public
    class function GetLetters(Captcha: TCaptcha): TArrayofLetter; override;
  end;

implementation

uses untLogger;

{ TNtLdN }

function IsRGB(const Color: Integer): Boolean;
begin
  Result := GetRGBDistance(Color) > 70;
end;

procedure SetDotsInDigits(var Captcha: TCaptcha);
var
  Grid: TGrid;
  X, Y, W: Integer;
begin
  Grid := TPixelGrid.GetGridCopy(Captcha.Grid);
  for X := 1 to Captcha.Width - 2 do
    for Y := 1 to Captcha.Height - 2 do
      if IsRGB(Captcha.Grid[X][Y]) then
      begin
        if (Captcha.Grid[X + 1][Y] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X + 1][Y])) then
          W := 1
        else
          W := 0;

        if (Captcha.Grid[X + 1][Y + 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X + 1][Y + 1])) then
          Inc(W);
        if (Captcha.Grid[X][Y + 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X][Y + 1])) then
          Inc(W);
        if (Captcha.Grid[X - 1][Y + 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X - 1][Y + 1])) then
          Inc(W);
        if (Captcha.Grid[X - 1][Y - 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X - 1][Y - 1])) then
          Inc(W);
        if (Captcha.Grid[X + 1][Y - 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X + 1][Y - 1])) then
          Inc(W);
        if (Captcha.Grid[X - 1][Y] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X - 1][Y])) then
          Inc(W);
        if (Captcha.Grid[X][Y - 1] <> $FFFFFF) and (not IsRGB(Captcha.Grid[X][Y - 1])) then
          Inc(W);

        if W > 4 then
          Grid[X][Y] := 0;
      end;
  Captcha.Grid := Grid;
end;

procedure ToBlack(var Captcha: TCaptcha);
var
  X, Y: Integer;
begin
  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
      if Captcha.Grid[X][Y] <> $FFFFFF then
        Captcha.Grid[X][Y] := 0;
end;

class function TNtLdN.GetLetters(Captcha: TCaptcha): TArrayofLetter;
var
  L: TLetter;
begin
  SetDotsInDigits(Captcha);
  Captcha.CleanByRGBDistance(1, 70);
  ToBlack(Captcha);
  Captcha.RemoveSmallObjects(0.5, 0.5, 25);
  Result := TEasyCaptcha.GetLetters(Captcha);

  for L in Result do
    L.ResizeToHeight(15);
end;

initialization
  // Achtung: Hier case sensitve bei SpecialClasses.ContainsKey(ClassName), daher alles lower case
  SpecialClasses.Add('ntldn.getletters', TNtLdN);

end.
