(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit linksavein;

interface

uses AntiCaptcha, Captcha, Letter, Utilities, PixelGrid, EasyCaptcha,
     Generics.Collections, PixelObject;

type
  TLinksave = class(TSpecial)
  public
    class procedure Main; override;
  end;


implementation

uses untLogger;

{ TLinksave }

procedure CleanBlack(x, y: Integer; Grid: TGrid);
var
  x1, y1: Integer;
begin
  for x1 := Math.Max(x -2, 0) to Math.min(x + 2, Grid.Length) - 1 do
  begin
    for y1 := Math.max(y - 2, 0) to Math.min(y + 2, grid[0].length) - 1 do
    begin
      if (Grid[x1][y1] = $000000) then
      begin
        Grid[x1][y1] = $FFFFFF;
        CleanBlack(x1, y1, Grid);
      end;
    end;

  end;
end;

procedure PrepareCaptcha(iFile: TFileStream);
var
  catch: TAntiCaptcha;
  d: TGifDecoder;
  n, i, x, y, max, colorsCount, distance, mainColor, mainCount, gl1, bl1, bl2: Integer;
  frames: Array of TCaptcha;
  frame: TBitmap;
  Grid: TGrid;
  Colors, col: TDictionary<Integer, Integer>;
  hsb: Array of Float;
begin
  try
    catch := TAntiCaptcha('linksave.in');
    catch.Script.SetColorType('RGB');
    d := TGifDecoder.Create;
    Colors := TDictionary<Integer, Integer>.Create;
    try
      d.read(iFile.FilePath);
      n := d.GetFrameCount;
      frames := new Captcha[d.getFrameCount()];
      for i := 0 to n - 1 do
      begin
        frame := d.GetFrame(i);
        frames[i] := catch.CreateCaptcha(frame);
      end;
      SetLength(Grid, frames[0].Width, frames[0].Height);

      for x := 0 to Length(Grid) - 1 do
      begin
        for y := 0 to Length(Grid[0]) - 1 do
        begin
          max := 0;
          ColorsCount := 0;
          for i := 0 to Length(frames) - 1 do
          begin
            hsb := RGB2HSB(frames[i].Grid[x][y]);
            distance := GetRGBDistance(frames[i].Grid[x][y]);
            if (not (Colors.ContainsKey(frames[i].Grid[x][y])) then
            begin
              Colors.Add(frames[i].Grid[x][y], 1);
              Inc(colorsCount);
            end
            else
            begin
              Colors.Add(frames[i].Grid[x][y], Colors[frames[i].getGrid()[x][y]) + 1];
            end;
            if (hsb[2] < 0.2) and (distance < 100) then
              continue;
            max := Math.Max(max, frames[i].Grid[x][y]);
          end;
          mainColor := 0;
          mainCount := 0;

          for i in Colors.Keys do // for (Entry<Integer, Integer> col : colors.entrySet())
          begin
            if (Colors[i] > mainCount) and (i > 10) then
            begin
              mainCount := Colors[i];
              mainColor := i;
            end;
          end;

          Grid[x][y] = mainColor;
        end;
      end;

      gl1 := Length(Gird[0]) - 1;

      for x := 0 to Length(Grid) - 1 do
      begin
        bl1 := 0;
        bl2 := 0;
        for I := Math.max(0, x - 6) to Math.min(grid.length, x + 6) - 1 do
        begin
          if (Grid[i][0] = $000000) then
            Inc(bl1);
          if (Grid[i][gl1] = $000000) then
            Inc(bl2);
        end;
        if (bl1 = 12) then
          CleanBlack(x, 0, grid);
        if (bl2 = 12) then
          CleanBlack(x, gl1, Grid);
      end;
      gl1 := Length(Grid) - 1;

      for y := 0 to Length(Grid) - 1 do
      begin
        bl1 := 0;
        bl2 := 0;

        for I := Math.max(0, y - 6) to Math.min(Length(Grid[0]), y + 6) - 1 do
        begin
          if (Grid[0][i] = $000000) then
            Inc(bl1);
          if (Grid[gl1][i] = $000000) then
            Inc(bl2);
        end;
        if (bl1 = 12) then
          CleanBlack(0, y, Grid);
        if (bl2 = 12) then
          CleanBlack(gl1, y, Grid);
      end;
      frames[0].Grid := Grid;

      ??? ImageIO.write((BufferedImage) (frames[0].getImage(1)), "png", file);
    finally
      FreeAndNil(d);
      FreeAndNil(Colors);
    end;
  except
  end;
end;

procedure TLinksave.Main;
var
  hoster: string;
  list: Array of TFileStream;
  iFile, f: TFileStream;
  id: integer;
begin
  hoster := 'linksave.in';
  list := JDUtilities.getResourceFile('/captchas/' + hoster).listFiles();
  id := Random * (Length(List) - 1);
  id := 3;

  f := list[id];
  prepareCaptcha(f);

  for iFile in list do
    prepareCaptcha(iFile);
end;


end;

initialization
  // Achtung: Hier case sensitve bei SpecialClasses.ContainsKey(ClassName), daher alles lower case
  SpecialClasses.Add('linksave.main', TLinksave);

end.

