(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit sharelinksbiz;

interface

uses AntiCaptcha, Captcha, Letter, Utilities, PixelGrid, EasyCaptcha,
     Generics.Defaults, Generics.Collections, PixelObject;

const
  BGColors: array[0..18] of Integer = ($746a5c, $847e5c, $847a64, $6c6244, $6c6654,
                                       $84765c, $8c7e58, $8c8274, $8c8264, $645e4c,
                                       $6c6249, $645e44, $7c725e, $746e54, $94825c,
                                       $7c6e50, $74664c, $94866c, $8c7e64);

type
  TShrLnksBz = class(TSpecial)
  public
    class function GetLetters(Captcha: TCaptcha): TArrayofLetter; override;
  end;

implementation

uses untLogger;

{ TShrLnksBz }

function Check(const Captcha: TPixelGrid; const X, Y: Integer): Boolean;
begin
  if (X >= 0) and (Y >= 0) and (Captcha.Width > X) and (Captcha.Height > Y) then
    if (Captcha.Grid[X][Y] = $FFFFFF) or (GetRGBColorDifference3($645E44, Captcha.Grid[X][Y]) > 10) then
    begin
      Captcha.Grid[X][Y] := $FFFFFF;
      Result := False;
    end
    else
      Result := True;
end;

function Quad(var Captcha: TPixelGrid; const X, Y, Distance: Integer): Integer;
var
  X1, Y1: Integer;
begin
  Result := 0;
  // if (x >= 0) and (y >= 0) and (captcha.getWidth() > x) and (captcha.getHeight() > y) and (captcha.grid[x][y] >= 0) and not (captcha.grid[x][y] = $ffffff) then
  // begin
  for X1 := -Distance to Distance do
    for Y1 := -Distance to Distance do
      if Check(Captcha, X + X1, Y + Y1) then
        Inc(Result);
  // end;
end;

(*
  * Overwrite the colored dots in digits with black dots
  * Die bunten Punkte in den Zahlen werden mit schwarzen Punkten ersetzt
*)
procedure SetDotsInDigits(var Captcha: TCaptcha; var Grid2: TGrid);
var
  Grid: TGrid;
  X, Y, W, Dist: Integer;
begin
  Dist := 4;
  Grid := TPixelGrid.GetGridCopy(Captcha.Grid);
  for X := 1 to Captcha.Width - 2 do
    for Y := 1 to Captcha.Height - 2 do
      if (Captcha.Grid[X][Y] = $FFFFFF) and (GetRGBColorDifference3($8C7E64, Grid2[X][Y]) > Dist) and (GetRGBColorDifference3($645E44, Grid2[X][Y]) > 10) then
      begin
        if (Captcha.Grid[X + 1][Y] = $FFFFFF) then
          W := 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X + 1][Y]) > Dist) then
            Inc(W);

        if not (Captcha.grid[X + 1][Y + 1] = $FFFFFF) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X + 1][Y + 1]) > Dist) then
            Inc(W);
        
        if not (Captcha.grid[X][Y + 1] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X][Y + 1]) > Dist) then
            Inc(W);

        if not (Captcha.grid[X - 1][Y + 1] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X - 1][Y + 1]) > Dist) then
            Inc(W);
        
        if not (Captcha.grid[X - 1][Y - 1] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X - 1][Y - 1]) > Dist) then
            Inc(W);

        if not (Captcha.grid[X + 1][Y - 1] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X + 1][Y - 1]) > Dist) then
            Inc(W);

        if not (Captcha.grid[X - 1][Y] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X + 1][Y + 1]) > Dist) then
            Inc(W);

        if not (Captcha.grid[X][Y - 1] = $ffffff) then
          W := W + 6
        else
          if (GetRGBColorDifference3($8C7E64, Grid2[X][Y - 1]) > Dist) then
            Inc(W);

        if W > 20 then
          Grid[X][Y] := $645E4C;
      end;

  Captcha.Grid := Grid;
end;

function Follow2(var Captcha: TPixelGrid; const X, Y, Color, LastColor, DistanceX, DistanceY, DistX2, DistY2: Integer; const LCSensitive: Boolean): Integer;
var
  LC, X1, Y1: Integer;
begin
  Result := 0;
  if (X >= 0) and (Y >= 0) and (Captcha.Width > X) and (Captcha.Height > Y) and (Captcha.Grid[X][Y] >= 0) and (Captcha.Grid[X][Y] = $FFFFFF) and ((not LCSensitive) or (Captcha.Grid[X][Y] = LastColor)) then
  begin
    LC := Captcha.Grid[X][Y];
    Captcha.Grid[X][Y] := Color;
    Inc(Result);

    for X1 := -DistX2 to DistX2 do
      for Y1 := -DistY2 to DistY2 do
        Result := Result + Follow2(Captcha, X + X1, Y + Y1, Color, LC, DistanceX, DistanceY, DistX2, DistY2, (X1 > DistanceX) or (Y1 > DistanceY) or (X1 < -DistanceX) or (Y1 < -DistanceY));
  end;
end;

procedure ToBlack(var Captcha: TCaptcha);
var
  X, Y: Integer;
begin
  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
      if Captcha.Grid[X][Y] <> $FFFFFF then
        Captcha.Grid[X][Y] := $000000;
end;

function GetFastObjects(Captcha: TPixelGrid; xDistance, yDistance, DistX2, DistY2, MinAnzahl: Integer): TList<TPixelObject>;
var
  i, X, Y: integer;
  PixelObject, Pix: TPixelObject;
  Next: Boolean;
begin
  i := -2;

  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
      if Captcha.Grid[X][Y] <> $FFFFFF then
      begin
        Follow2(Captcha, X, Y, i, Captcha.Grid[X][Y], xDistance, yDistance, DistX2, DistY2, False);
        Dec(i);
      end;

  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
      if Captcha.Grid[X][Y] <> $FFFFFF then
        Captcha.Grid[X][Y] := Captcha.Grid[X][Y] * -300;

  Result := TList<TPixelObject>.Create(TComparer<TPixelObject>.Construct(PixelObjectComparer));
  Next := False;
  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
    begin
      Next := False;
      if Captcha.Grid[X][Y] <> $FFFFFF then
      begin
        for PixelObject in Result do
        begin
          if Next then
            Continue;

          if PixelObject.Average = Captcha.Grid[X][Y] then
          begin
            PixelObject.Add(X, Y, Captcha.Grid[X][Y]);
            Next := True;
          end;
        end;

        if Next then
          Continue;

        Pix := TPixelObject.Create(Captcha);
        Pix.Add(X, Y, Captcha.Grid[X][Y]);
        Result.Add(Pix);
      end;
    end;

  for PixelObject in Result do
    if PixelObject.Size <= MinAnzahl then
      Result.Remove(PixelObject);

  Result.Sort;
end;

(*function GetBigLetters(Captcha: TCaptcha): TArrayOfLetter;
var
  Grid: TGrid;
  CoLetters: TList<TPixelObject>;
  X, Y, i: Integer;
  Let: TLetter;
  pixelObject: TPixelObject;
begin
  (Captcha.Owner as TAntiCaptcha).LoadMTHFile('lettersbig.mth');
  Grid := TPixelGrid.getGridCopy(Captcha.Grid);  
  for X := 0 to Captcha.Width - 1 do
    for Y := 0 to Captcha.Height - 1 do
      if Captcha.Grid[X][Y] <> $FFFFFF then
        if Quad(Captcha, X, Y, 1) < 5 then
          Captcha.Grid[X][Y] := $FFFFFF;

  SetDotsInDigits(Captcha, Grid);
  SetDotsInDigits(Captcha, Grid);
  SetDotsInDigits(Captcha, Grid);
  SetDotsInDigits(Captcha, Grid);
  SetDotsInDigits(Captcha, Grid);
  Captcha.ResizeToHeight(100);
  CoLetters := GetFastObjects(Captcha, 1, 3, 1, 3, 200);

  SetLength(Result, CoLetters.Count);
  i := 0;
  for PixelObject in CoLetters do
  begin
    Let := TLetter(PixelObject.ToLetter);
    ToBlack(Let);
    Let.RemoveSmallObjects(0.9, 0.9, 4, 2, 2);
    Let.Clean;
    Let.ResizeToHeight(20);
    Result[i] := Let;
    Inc(i);
  end;
  Captcha.Grid := Grid;
  Result := TEasyCaptcha.LetterFilter(Result, Captcha.owner);
  (Captcha.Owner as TAntiCaptcha).LoadMTHFile;
  if (Length(Result) > 0) and (Result[0].DecodedValue = '1') then
    Result[0].DecodedValue := 'i';
end;

function GetSmallLetters(Captcha: TCaptcha): TList<TList<TPixelObject>>;
var
  Grid: TGrid;
  x, y, col, i, minY, minX: integer;
  coLetters: TList<TPixelObject>;
  PixelObject, arrayList: TPixelObject;

Label Outer;
begin
  grid = TPixelGrid.GetGridCopy(Captcha.Grid);
  for x := 0 to Captcha.Width - 1 do
  begin
    for y := 0 to Captcha.Height - 1 do
    begin
      for col in bgColors do
      begin
        if (getRGBColorDifference3(col, Captcha.Grid[x][y]) < 15) or (getRGBColorDifference3(col, Captcha.Grid[x][y]) < 15) then
          Captcha.Grid[x][y] = $FFFFFF;
      end;
    end;
  end;

  coLetters := getFastObjects(captcha, 2, 1, 9, 2, 8);
  i := 0;

  Result := TList< TList<TPixelObject> >.Create;

  while (TPixelObject(coLetters).Size > 0) do
  begin
    Result.Add(TList<TpixelObject>);
    minY := MAXINT;
    for PixelObject in coLetters do
    begin
      if (PixelObject.YMax < minY) then
        minY := PixelObject.YMax;
    end;

    for (Iterator<PixelObject> iterator = coLetters.iterator(); iterator.hasNext();)
    begin
      PixelObject := (PixelObject) iterator.next();
      if (Abs(PixelObject.YMax - minY) < 10) then
      begin
        Result.Get(i).Add(PixelObject);
        iterator.remove();
      end;
    end;

    Goto outer;
    for (Iterator<PixelObject> iterator = coLetters.iterator(); iterator.hasNext();)
    begin
      PixelObject pixelObject = (PixelObject) iterator.next();
      if (Math.abs(pixelObject.getYMin() - minY) < 20)
      begin
        for (PixelObject arrayList : coLetters2.get(i))
        begin
          if (Math.abs(pixelObject.getXMin() - arrayList.getXMin()) < 4) then
            outer:
        end;
        Result.get(i).add(PixelObject);
        iterator.remove();
      end;
    end;
    inc(i);
  end;

  for arrayList in Result do
  begin
    Collections.Sort(arrayList);
  end;
  Captcha.Grid = Grid;
end;      *)


class function TShrLnksBz.GetLetters(Captcha: TCaptcha): TArrayofLetter;
(*var
  letbig, letters: TArrayOfLetter;
  letter, let, let2: TLetter;
  tx, num, numnew: string;
  numorg, b, number: integer;
  i, last, x, y: integer;
  c: char;
  letterList: TList< TList<TPixelObject> >;
  lastObje, pixelObject, PixelObject2: TPixelObject;
  lets3: TList<TLetter>;     *)
begin
    (*    letbig := getBigLetters(Captcha);
        tx = "";
        for letter in letbig do
        begin
          if not (letter.DecodedValue = null) then
            tx := tx + letter.DecodedValue;
        end;
        tx := tx.substring(1);
        numorg := StringToInteger(tx);
        b := 0;
        c := letbig[0].DecodedValue.charAt(0);
        for i = Ord('a') to (Ord('z')- 1) do
        begin
          if Chr(i) = c then
            inc(b);
        end;
        (*
        for (char i = 'a'; i < 'z'; i++) {
            if (i == c) break;
            b++;
        }
        *)     (*

        letterList := GetSmallLetters(Captcha);
        last := -1;
        lastObje := nil;
        x := -1;
        y := -1;

        for pixelObject in letterList.get(b) do
        begin
          try
            if (PixelObject.Size > 8) then
            begin
              let := PixelObject.toColoredLetter;
              toBlack(let);
              lets3 := TList<TLetter>.Create;

              lets2 := getFastObjects(let, 1, 1, 1, 1, 1);
              for PixelObject2 in lets2 do
              begin
                if (PixelObject2.Size > 5) then
                begin
                  let2 := PixelObject2.toColoredLetter;
                  toBlack(let2);
                  let2.autoAlign;
                  let2.ReSizeToHeight(30);
                  lets3.Add(let2);
                end;
              end;

              number := -1;
              letters := lets3.toArray(new Letter[] {});
              letters = EasyCaptcha.letterFilter(letters, captcha.owner);

              if (lets3.Size > 0) then
              begin
                if (letters[0].DecodedValue.EndsWith(letbig[0].DecodedValue)) then
                begin
                  num := '';
                  for letter in letters do
                  begin
                    if (not (letter.DecodedValue = null)) then
                      num := num + letter.DecodedValue;
                  end;
                  num := num.substring(1);

                  if (Length(num) > 0) then
                  begin
                    if (Length(num) > 1) and (num[0] = 'z') then
                      num := num.substring(1);
                    if (Length(num) > 0) then
                    begin
                      numnew := num.replaceAll("[^0-9]", "1");
                      number := StrToInt(numnew);
                    end;
                  end;
                end;
              end;
              if (number >= 0) and (number > (last - 8)) and (number < (last + 12)) then
                last := number
              else
              begin
                inc(last);
                number := last;
              end;
              if (numorg = number) then
              begin
                x = let.getLocation()[0] + (let.getWidth() / 2);
                y = let.getLocation()[1] + (let.getHeight() / 2);
                break;
              end
              else
                if (number > numorg) and (not (lastObje = null)) then
                begin
                  x := lastObje.getLocation()[0] + (lastObje.getWidth() / 2);
                  y := lastObje.getLocation()[1] + (lastObje.getHeight() / 2);
                  break;
                end;

              lastObje := pixelObject;
            end;
          except
          end;
        end;        *)
end;

initialization
  // Achtung: Hier case sensitve bei SpecialClasses.ContainsKey(ClassName), daher alles lower case
  SpecialClasses.Add('shrlnksbz.getletters', TShrLnksBz);

end.
