(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit CSTools;

interface

uses SysUtils, Classes, Windows, StrUtils, Forms;

type
  TArrayofstring = array of string;

  TExecuteWaitEvent = procedure(const ProcessInfo: TProcessInformation;
                                  var ATerminate: Boolean; var MaxWait: Integer);

function LoadStringFromFile(Filename: string): string;
function FindInString(Text, SearchFrom, SearchTo: string; FirstOnly: Boolean; var Return: TArrayofstring): Boolean;
function GetTempDir: string;
function Explode(const Separator, s: string; Limit: Integer = 0): TArrayofstring;
function IsFileInUse(Path: string): Boolean;
function GetFileSize(const szFile: string): Int64;
function StrCount(s: string; c: Char): Integer;
procedure ExecuteFile(const AFilename: string;
                 ShowCMD: DWORD; AParameter, ACurrentDir: string; AWait: Boolean;
                 AOnWaitProc: TExecuteWaitEvent = nil; MaxWait: Integer = 0);

implementation

function LoadStringFromFile(Filename: string): string;
var
  b: Byte;
  i: Integer;
  FS: TFileStream;
begin
  FS := TFileStream.Create(Filename, fmOpenRead or fmShareDenyNone);
  try
    SetLength(Result, FS.Size);
    if FS.Size = 0 then
      Exit;

    for i := 1 to FS.Size do
    begin
      FS.Read(b, 1);
      Result[i] := Chr(b);
    end;
  finally
    FS.Free;
  end;
end;

function FindInString(Text, SearchFrom, SearchTo: string; FirstOnly: Boolean; var Return: TArrayofstring): Boolean;
var
  i: Integer;
  FoundString: string;
begin
  Result := False;
  SetLength(Return, 0);
  i := Pos(SearchFrom, Text);
  while i > 0 do
  begin
    Result := True;
    i := i + Length(SearchFrom);
    FoundString := Copy(Text, i, PosEx(SearchTo,Text,i) - i);
    SetLength(Return, length(Return) + 1);
    Return[high(Return)] := FoundString;
    Delete(Text, 1, i);
    i := Pos(SearchFrom, Text);
    if FirstOnly then Break;
  end;
end;

function Explode(const Separator, s: string; Limit: Integer = 0): TArrayofstring;
var
  SepLen: Integer;
  F, P: PChar;
  ALen, Index: Integer;
begin
  SetLength(Result, 0);
  if (S = '') or (Limit < 0) then Exit;
  if Separator = '' then
  begin
    SetLength(Result, 1);
    Result[0] := S;
    Exit;
  end;
  SepLen := Length(Separator);
  ALen := Limit;
  SetLength(Result, ALen);


  Index := 0;
  P := PChar(S);
  while P^ <> #0 do
  begin
    F := P;
    P := StrPos(P, PChar(Separator));
    if (P = nil) or ((Limit > 0) and (Index = Limit - 1)) then P := StrEnd(F);
    if Index >= ALen then
    begin
      Inc(ALen, 5); // mehrere auf einmal um schneller arbeiten zu knnen
      SetLength(Result, ALen);
    end;
    SetString(Result[Index], F, P - F);
    Inc(Index);
    if P^ <> #0 then Inc(P, SepLen);
  end;
  if Index < ALen then SetLength(Result, Index); // wirkliche Lnge festlegen
end;

function GetTempDir: string;
var
   lng: DWORD;
begin
  SetLength(Result, MAX_PATH);
  lng := GetTempPath(MAX_PATH, PChar(Result));
  SetLength(Result, lng);
end;

function IsFileInUse(Path: string): Boolean;
var
  hFile: THandle;
begin
  Result := False;
  if not FileExists(Path) then
    Exit;
  hFile := CreateFile(pchar(Path), GENERIC_READ or GENERIC_WRITE or GENERIC_EXECUTE,
                      0, nil, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
  Result := hFile = INVALID_HANDLE_VALUE;
  if not Result then
    CloseHandle(hFile);
end;

function GetFileSize(const szFile: string): Int64;
var
  fFile: THandle;
  wfd: TWIN32FINDDATA;
begin
  Result := 0;
  if not FileExists(szFile) then
    Exit;

  fFile := FindFirstfile(pchar(szFile),wfd);
  if fFile = INVALID_HANDLE_VALUE then
    Exit;

  Result := (wfd.nFileSizeHigh * MAXDWORD) + wfd.nFileSizeLow;
  Windows.FindClose(fFile);
end;

function StrCount(s: string; c: Char): Integer;
var
  i: Integer;
begin
  Result := 0;

  for i := 1 to Length(s) do
    if s[i] = c then
      Inc(Result);
end;

procedure ExecuteFile(const AFilename: string;
                 ShowCMD: DWORD; AParameter, ACurrentDir: string; AWait: Boolean;
                 AOnWaitProc: TExecuteWaitEvent = nil; MaxWait: Integer = 0);
var
  si: TStartupInfo;
  pi: TProcessInformation;
  bTerminate: Boolean;
begin
  bTerminate := False;

  if Length(ACurrentDir) = 0 then
    ACurrentDir := ExtractFilePath(AFilename);

  if AnsiLastChar(ACurrentDir) = '\' then
    Delete(ACurrentDir, Length(ACurrentDir), 1);

  FillChar(si, SizeOf(si), 0);
  with si do begin
    cb := SizeOf(si);
    dwFlags := STARTF_USESHOWWINDOW;
    wShowWindow := showcmd;
  end;

  FillChar(pi, SizeOf(pi), 0);
  AParameter := Format('"%s" %s', [AFilename, TrimRight(AParameter)]);

  if CreateProcess(Nil, PChar(AParameter), Nil, Nil, False,
                   CREATE_DEFAULT_ERROR_MODE or CREATE_NEW_CONSOLE or
                   NORMAL_PRIORITY_CLASS, Nil, PChar(ACurrentDir), si, pi) then
  try
    if AWait then
      while WaitForSingleObject(pi.hProcess, 50) <> Wait_Object_0 do
      begin
        if Assigned(AOnWaitProc) then
        begin
          AOnWaitProc(pi, bTerminate, MaxWait);
          if bTerminate then
            TerminateProcess(pi.hProcess, Cardinal(-1));
        end;

        Application.ProcessMessages;
      end;
  finally
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
  end;
end;

end.
