﻿(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit Utilities;

interface

uses Windows, CSTools, Math;

const
  MININT: Integer = -2147483648;

type
  TArrayofInteger = array of Integer;
  TArrayofDouble = array of Double;
  T2DArrayofInteger = array of TArrayofInteger;
  TGrid = T2DArrayofInteger;
  TBooleanMap = array [0..2] of array [0..2] of Boolean;

function RGBToHSB(RGB: TArrayofInteger): TArrayofDouble;
function RGBToLAB(R, G, B: Integer): TArrayofInteger;
function HexToRGB(Value: Integer): TArrayofInteger;
function RgbToHex(Value: TArrayofInteger): Integer;
function MixColors(A, B, ga, gb: Integer): Integer; overload;
function MixColors(A, B: Integer): Integer; overload;
function TurnCoordinates(X, Y, nullX, nullY: Integer; Angle: Double): TArrayofInteger;

function GetHsbColorDifference(rgbA, rgbB: TArrayofInteger): Double;
function GetRGB(A: Integer): TArrayofInteger;
function GetColorDifference(A, B: Integer): Double; overload;
function GetColorDifference(rgbA, rgbB: TArrayofInteger): Double; overload;
function GetBrightnessColorDifference(A, B: Integer): Double; overload;
function GetBrightnessColorDifference(RGBA, RGBB: TArrayofInteger): Double; overload;
function GetRGBDistance(const Color: Integer): Integer;
function GetRGBColorDifference2(const Color, Color2: Integer): Double;
function GetRGBColorDifference3(const Color, Color2: Integer): Double;

function GetJumperStart(iFrom, iTo: Integer): Integer;
function CheckJumper(X, iFrom, iTo: Integer): Boolean;
function NextJump(X, iFrom, iTo, Step: Integer): Integer;

implementation

// Es dumme Delphi hat natürlich mal wieder nich alle Math-Funtkionen :rolleyes:
function atan2(y: Extended; x: Extended): Extended; assembler;
asm
  fld [y]
  fld [x]
  fpatan
end;

// Und auch an den Color Funktionen scheitert es
function RGBToHSB(RGB: TArrayofInteger): TArrayofDouble;
var
  minRGB, maxRGB, delta: Double;
  h , s , b: Double;
begin
  H := 0.0 ;
  minRGB := Math.Min(Math.Min(rgb[0], rgb[1]), rgb[2]);
  maxRGB := Math.Max(Math.Max(rgb[0], rgb[1]), rgb[2]);
  delta := (maxRGB - minRGB);
  b := maxRGB;
  if (maxRGB <> 0.0) then
    s := 255.0 * Delta / maxRGB
  else
    s := 0.0;

  if (s <> 0.0) then
  begin
    if rgb[0] = maxRGB then
      h := (rgb[1] - rgb[2]) / Delta
    else
      if rgb[1] = maxRGB then
        h := 2.0 + (rgb[2] - rgb[0]) / Delta
      else
        if rgb[2] = maxRGB then
          h := 4.0 + (rgb[0] - rgb[1]) / Delta
  end
  else
    h := -1.0;

  h := h * 60 ;
  if h < 0.0 then
    h := h + 360.0;

  SetLength(Result, 3);
  Result[0] := h;
  Result[1] := s * 100 / 255;
  Result[2] := b * 100 / 255;
end;

function RGBToLAB(R, G, B: Integer): TArrayofInteger;
var
  r_, g_, b_, X, Y, Z, fx, fy, fz, xr, yr, zr: Extended;
  Ls, as_, bs: Extended;
  eps, k: Extended;
begin
  SetLength(Result, 3);

  eps := 216 / 24389;
  k := 24389 / 27;

  Xr := 0.964221; // reference white D50
  Yr := 1.0;
  Zr := 0.825211;

  // RGB to XYZ
  r_ := R / 255; // R 0..1
  g_ := G / 255; // G 0..1
  b_ := B / 255; // B 0..1

  // assuming sRGB (D65)
  if (r_ <= 0.04045) then
    r_ := r_ / 12
  else
    r_ := Power((r_ + 0.055) / 1.055, 2.4);

  if (g_ <= 0.04045) then
    g_ := g_ / 12
  else
    g_ := Power((g_ + 0.055) / 1.055, 2.4);

  if (b_ <= 0.04045) then
    b_ := b_ / 12
  else
    b_ := Power((b_ + 0.055) / 1.055, 2.4);

  X := 0.436052025 * r_ + 0.385081593 * g_ + 0.143087414 * b_;
  Y := 0.222491598 * r_ + 0.71688606 * g_ + 0.060621486 * b_;
  Z := 0.013929122 * r_ + 0.097097002 * g_ + 0.71418547 * b_;

  // XYZ to Lab
  xr := X / Xr;
  yr := Y / Yr;
  zr := Z / Zr;

  if (xr > eps) then
    fx := Power(xr, 1 / 3.0)
  else
    fx := ((k * xr + 16.0) / 116.0);

  if (yr > eps) then
    fy := Power(yr, 1 / 3.0)
  else
    fy := ((k * yr + 16.0) / 116.0);

  if (zr > eps) then
    fz := Power(zr, 1 / 3.0)
  else
    fz := ((k * zr + 16.0) / 116);

  Ls := 116 * fy - 16;
  as_ := 500 * (fx - fy);
  bs := 200 * (fy - fz);

  Result[0] := Trunc(2.55 * Ls + 0.5);
  Result[1] := Trunc(as_ + 0.5);
  Result[2] := Trunc(bs + 0.5);
end;


(*
  * Wandelt eine decimale hexzahl(0-256^3) in die 3 RGB Farbkomponenten um.
  *
  * @param value
  * @return RGB Werte
*)
function HexToRGB(Value: Integer): TArrayofInteger;
var
  v: TArrayofInteger;
begin
  SetLength(v, 4);
  v[0] := Value div 65536;
	v[1] := (Value - Value div 65536 * 65536) div 256;
	v[2] :=	Value - Value div 65536 * 65536 - (Value - Value div 65536 * 65536) div 256 * 256;
	v[3] :=	0;
	Result := v;
end;

(*
  * Wandelt ein RGB Array in die zugehöroge decimale hexzahl um.
	*
	* @param value
	* @return 32 BIt Farbwert
*)
function RGBToHex(Value: TArrayofInteger): Integer;
begin
	Result := Value[0] * 65536 + Value[1] * 256 + Value[2];
end;

(*
	* Mischt zwei dezimal zahlen. dabei werden ga und gb als
	* gewichtungsfaktoren verwendet mixColors(0xff0000,0x00ff00,3,1) mischt 3
	* Teile Rot und einen teil grün
	*
	* @param a
	* @param b
	* @param ga
	* @param gb
  * @return 32 Bit Farbwert
*)
function MixColors(A, B, ga, gb: Integer): Integer;
var
  av, bv, res: TArrayofInteger;
  R, G, Bl: Integer;
begin
  av := hexToRgb(a);
	bv := hexToRgb(b);
	R := (av[0] * ga + bv[0] * gb) div (ga + gb);
 	G := (av[1] * ga + bv[1] * gb) div (ga + gb);
 	Bl := (av[2] * ga + bv[2] * gb) div (ga + gb);

  SetLength(res, 3);
  res[0] := R;
  res[1] := G;
  res[2] := Bl;

 	Result := RGBToHex(res);
end;

(*
 * Mischt zwei Fraben 1:1
 *
 * @param a
 * @param b
 * @return 32 Bit Farbwert
 *)
function MixColors(A, B: Integer): Integer;
var
  av, bv, ret: TArrayofInteger;
begin
  av := hexToRgb(a);
  bv := hexToRgb(b);
  SetLength(ret, 3);
  ret[0] := (av[0] + bv[0]) div 2;
  ret[1] := (av[1] + bv[1]) div 2;
  ret[2] := (av[2] + bv[2]) div 2;
  Result := RGBToHex(ret);
end;

	(**
	 * Dreht die Koordinaten x und y um den Mittelpunkt nullX und nullY um den
	 * Winkel winkel
	 *
	 * @param x
	 * @param y
	 * @param nullX
	 * @param nullY
	 * @param winkel
	 * @return neue Koordinaten
	 *)
function TurnCoordinates(X, Y, nullX, nullY: Integer; Angle: Double): TArrayofInteger;
var
  newX, newY, yTrans, xTrans: Integer;
  aktAngle, Radius: Double;
begin
  Angle := Angle / 180.0;
  newX := X - nullX;
	newY := Y - nullY;
	aktAngle := atan2(newY, newX);

  SetLength(Result, 2);
  Radius := Sqrt(newX * newX + newY * newY);
  yTrans := Round(Radius * Sin((aktAngle + Angle * Pi)));
  xTrans := Round(Radius * Cos((aktAngle + Angle * Pi)));
  Result[0] := xTrans + nullX;
  Result[1] := yTrans + nullY;
end;

function GetHsbColorDifference(rgbA, rgbB: TArrayofInteger): Double;
var
  hsbA, hsbB: Double;
begin
  hsbA := rgbTohsb(rgbA)[0] * 100;
	hsbB := rgbTohsb(rgbB)[0] * 100;
	Result := Abs(hsbA - hsbB);
end;

function GetRGB(A: Integer): TArrayofInteger;
begin
  SetLength(Result, 3);
  Result[0] := GetRValue(a);
  Result[1] := GetGValue(a);
  Result[2] := GetBValue(a);
end;

function GetColorDifference(A, B: Integer): Double;
begin
  Result := GetColorDifference(GetRGB(a), GetRGB(b));
end;

function GetColorDifference(rgbA: TArrayofInteger; rgbB: TArrayofInteger): Double;
var
  labA, labB: TArrayofInteger;
  dif0, dif1, dif2: Integer;
begin
  labA := RGBToLAB(rgbA[0], rgbA[1], rgbA[2]);
  labB := RGBToLAB(rgbB[0], rgbB[1], rgbB[2]);
  dif0 := labA[0] - labB[0];
  dif1 := labA[1] - labB[1];
  dif2 := labA[2] - labB[2];
  Result := Sqrt(dif0 * dif0 + dif1 * dif1 + dif2 * dif2);
end;

function GetBrightnessColorDifference(A, B: Integer): Double;
begin
  if A = B then
    Exit(0);

  Result := GetBrightnessColorDifference(GetRGB(A), GetRGB(B));
end;

function GetBrightnessColorDifference(RGBA, RGBB: TArrayofInteger): Double;
var
  HSBA, HSBB: Single;
begin
  HSBA := RGBToHSB(RGBA)[2] * 100;
  HSBB := RGBToHSB(RGBB)[2] * 100;

  Result := Abs(HSBA - HSBB);
end;

function GetRGBDistance(const Color: Integer): Integer;
var
  br, bg, rg: Integer;
begin
  br := Abs(GetBValue(Color) - GetRValue(Color));
  bg := Abs(GetBValue(Color) - GetGValue(Color));
  rg := Abs(GetGValue(Color) - GetRValue(Color));

  Result := (br + bg + rg) div 3;
end;

function GetRGBColorDifference2(const Color, Color2: Integer): Double;
var
  Dif: array[0..2] of Integer;
begin
  if Color = Color2 then
    Exit(0);

  Dif[0] := GetRValue(Color) - GetRValue(Color2);
  Dif[1] := GetGValue(Color) - GetGValue(Color2);
  Dif[2] := GetBValue(Color) - GetBValue(Color2);
  Result := (Abs(Dif[0]) + Abs(Dif[1]) + Abs(Dif[2])) / 3;
end;

function GetRGBColorDifference3(const Color, Color2: Integer): Double;
var
  Dif: array[0..2] of Integer;
begin
  if Color = Color2 then
    Exit(0);

  Dif[0] := GetRValue(Color) - GetRValue(Color2);
  Dif[1] := GetGValue(Color) - GetGValue(Color2);
  Dif[2] := GetBValue(Color) - GetBValue(Color2);
  Result := Max(Abs(Dif[0]), Max(Abs(Dif[1]), Abs(Dif[2])));
end;

function GetJumperStart(iFrom, iTo: Integer): Integer;
begin
  Result := iFrom + Trunc((iTo - iFrom) / 2);
end;

function CheckJumper(X, iFrom, iTo: Integer): Boolean;
begin
  Result := (X >= iFrom) and (X <= iTo);
end;

function NextJump(X, iFrom, iTo, Step: Integer): Integer;
var
  Start, Dif: Integer;
begin
  Start := GetJumperStart(iFrom, iTo);
  if X = Start then
  begin
    Result := Start + Step;
    if Result > iTo then
      Result := Start - Step;
  end
  else
    if X > Start then
    begin
      Dif := X - Start;
      Result := Start - Dif;
    end
    else
    begin
      Dif := Start - X + Step;
      Result := Start + Dif;
      if Result > iTo then
        Result := Start - Dif;
    end;
end;



end.
