(*
     Catcher! OCR Module  -  Port of JAntiCaptcha
     Copyright (C) 2010  CandiSoft

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

unit untLogger;

interface

{$IFDEF DEBUG}uses StdCtrls;{$ENDIF}

type
  TMessageOutput = procedure(Msg: PWideChar); stdcall;

  TLogger = class
  private
    {$IFDEF DEBUG}FControl: TMemo;{$ENDIF}
    FOutput: TMessageOutput;

    procedure DoLog(Msg: string);
  public
    constructor Create;

    procedure Severe(Msg: string);
    procedure Fine(Msg: string);
    procedure Warning(Msg: string);
    procedure Info(Msg: string);
    procedure Finer(Msg: string);
    procedure Log(Msg: string);

    {$IFDEF DEBUG}property Control: TMemo read FControl write FControl;{$ENDIF}
    property Output: TMessageOutput read FOutput write FOutput;
  end;

var
  Logger: TLogger;

implementation

{ TLogger }

constructor TLogger.Create;
begin
  FOutput := nil;
end;

procedure TLogger.DoLog(Msg: string);
begin
  {$IFDEF DEBUG}
  if FControl <> nil then
    FControl.Lines.Add(Msg);
  {$ENDIF}
  if @FOutput <> nil then
    FOutput(PChar(Msg));
end;

procedure TLogger.Severe(Msg: string);
begin
  DoLog('[Severe]  ' + Msg);
end;

procedure TLogger.Fine(Msg: string);
begin
  DoLog('[Fine]    ' + Msg);
end;

procedure TLogger.Finer(Msg: string);
begin
  DoLog('[Finer]   ' + Msg);
end;

procedure TLogger.Warning(Msg: string);
begin
  DoLog('[Warning] ' + Msg);
end;

procedure TLogger.Info(Msg: string);
begin
  DoLog('[Info]    ' + Msg);
end;

procedure TLogger.Log(Msg: string);
begin
  DoLog('[Always]  ' + Msg);
end;

initialization
  Logger := TLogger.Create;

finalization
  Logger.Free;

end.
